# pyright: reportUndefinedVariable=false
#%% Load modules
import pandas as pd

#%% Importing blastp results
blastp_table_path = snakemake.input[0]
blastp_df = pd.read_csv(blastp_table_path, sep='\t')

#%%
if blastp_df.empty == False:

    #Extract motif information from subject information
    blastp_df[['REbase_id', 'motif_length']] = blastp_df['salltitles'].str.split('   ', expand = True)
    blastp_df[['motif', 'aa_length']] = blastp_df['motif_length'].str.split('  ', expand = True)
    blastp_df.drop(['motif_length', 'aa_length'], axis=1, inplace=True)


    # Filter blastp results based on identity and coverage parameters.
    qseq_groups = blastp_df.groupby('qseqid')
    idx_to_keep = []
    for name, group in qseq_groups:
        qseqs_sign = group.loc[(group['pident'] >= snakemake.params['identity']) & (group['qcovs'] >= snakemake.params['qcovs'])]
        if len(qseqs_sign) > 1:
            idx = qseqs_sign['bitscore'].idxmax()
            idx_to_keep.append(idx)
        elif len(qseqs_sign) == 1:
            idx_to_keep.append(qseqs_sign.index[0])
        else:
            pass

    blastp_df_sign = blastp_df.loc[idx_to_keep]

    # Export filterede blastp results
    blastp_df_sign.to_csv(snakemake.output[0], sep = "\t", index = False)

else:
    blastp_df.to_csv(snakemake.output[0], sep = "\t", index = False)



"""
Utility functions
"""
from itertools import product
import numpy as np
import re
import random
import nanomotif as nm
import os
np.random.seed(1)

def has_n_character_stretches_of_length_m(sequence, n, m, character):
    """
    Check if the given sequence has a segment of three or more consecutive dots.

    Parameters:
    sequence (str): The sequence to be checked.

    Returns:
    bool: True if there is a segment of three or more consecutive dots, False otherwise.
    """
    regex_str = rf"({character}){{{m},}}"
    matches = re.findall(regex_str, sequence)
    return len(matches) >= n

def motif_type(motif_str):
    if has_n_character_stretches_of_length_m(motif_str, 2, 2, "N"):
        return "ambiguous"
    elif re.search(r"(N){3,}", motif_str):
        return "bipartite"
    elif nm.seq.reverse_compliment(motif_str) == motif_str:
        return "palindrome"
    else:
        return "non-palindrome"

def generate_kmers(k: int, alphabet = ["A", "T", "C", "G"]):
    """
    Generate all possible k-mers of a given alphabet of sizes 1 to k
    """
    return [''.join(p) for i in range(1, k + 1) for p in product(alphabet, repeat=i)]
    


def subseq_indices(subseq, seq):
    """
    Find occourance indeces of a subseq in a seq
    """
    compiled_subseq = re.compile(subseq)
    index = [match.start() for match in re.finditer(compiled_subseq, seq)]
    return np.array(index)

def calculate_match_length(regex):
    # Remove non-capturing groups for simplicity
    regex = re.sub(r'\[.*\]', '.', regex)

    return len(regex)


def all_lengths_equal(iterator):
    """
    Checks whether the lengths of all elements in an iterable are equal.

    The function will return True even if the iterable is empty. It requires that the elements of the iterable
    also be iterable, such as strings, lists, and tuples.

    Args:
        iterator (Iterable): An iterable object containing other iterable elements.

    Returns:
        bool: True if all lengths are equal or if iterable is empty, False otherwise.

    Examples:
    >>> all_lengths_equal(['abc', 'def', 'ghi'])
    True
    >>> all_lengths_equal([[1, 2, 3], [4, 5, 6]])
    True
    >>> all_lengths_equal([])
    True
    >>> all_lengths_equal(['abc', 'de'])
    False
    """
    iterator = iter(iterator)
    try:
        first = len(next(iterator))
    except StopIteration:
        return True
    return all(first == len(x) for x in iterator)


def concatenate_motif_position_files(files, path):
    combined_data = {}
    for f in files:
        contig_name = os.path.basename(f).split("_")[0:-3]
        contig_name = "_".join(contig_name)
        mod_type = os.path.basename(f).split("_")[-3]

        with np.load(f, allow_pickle=True) as data:
            motif_data = {key: data[key].item() for key in data.files}
            
            if contig_name not in combined_data:
                combined_data[contig_name] = {}
            
            combined_data[contig_name][mod_type] = motif_data
    
    np.savez(os.path.join(path, "motif_positions_combined.npz"), **combined_data)

def clean_up_motif_positions_files(files):
    for f in files:
        os.remove(f)




import numpy as np
import polars as pl
import os
from polars import col
import logging as log
import itertools
from itertools import takewhile
from scipy.stats import entropy
import multiprocessing
from multiprocessing import get_context
import progressbar
from nanomotif.constants import *
from nanomotif.model import BetaBernoulliModel
from nanomotif.utils import subseq_indices, calculate_match_length
from nanomotif.seq import EqualLengthDNASet, DNAsequence, DNAarray
from nanomotif.candidate import Motif, MotifTree
from nanomotif.logger import configure_logger
from nanomotif.parallel import update_progress_bar
from nanomotif.seed import set_seed
import heapq as hq
import networkx as nx
import warnings
import time
from typing import Generator, Optional, List, Tuple

##########################################
# Motif candidate scoring
##########################################

def methylated_motif_occourances(
        motif, 
        sequence, 
        methylated_positions
    ) -> tuple:
    """
    Get occourances of a motif in a contig

    Parameters:
    - motif (str): The motif to search for.
    - seq (str): The contig to search in.
    - contig_meth_positions (list): The positions of the methylation sites in the contig.

    Returns:
    - tuple: A tuple of two numpy arrays, the first containing the positions of the methylated motifs, the second containing the positions of the non-methylated motifs.
    """
    assert len(motif) > 0, "Motif is empty"
    assert len(sequence) > 0, "Sequence is empty"
    assert type(motif) == Motif, "Motif is not a Motif type"
    # Get the index of the methylation in the motif in the contig
    motif_index = subseq_indices(motif.string, sequence) + motif.mod_position

    # Methylated motif positions
    meth_occurences = np.intersect1d(methylated_positions, motif_index)

    # Non-methylated motif positions
    nonmeth_occurences =  np.setdiff1d(motif_index, methylated_positions)

    return meth_occurences, nonmeth_occurences

def motif_model_contig(
        pileup, 
        contig: str, 
        motif, 
        save_motif_positions=False,
        na_positions = None
    ):
    """
    Get the posterior for a single motif. Uses number of methylated motifs as methylation count.

    Parameters:
    - pileup (Pileup): The pileup to be processed.
    - contig (str): The contig to be processed.
    - motif (str): The motif to be processed.
    - motif_meth_index (int): The index of the methylation site in the motif.

    Returns:
    - BetaBernoulliModel: The model for the candidate.
    """
    # Strip motif of periods    # Strip motif of periods
    motif_stripped = motif.new_stripped_motif()

    meth_positions_fwd = pileup.filter(pl.col("strand") == "+")["position"].to_numpy()
    meth_positions_rev = pileup.filter(pl.col("strand") == "-")["position"].to_numpy()

    index_meth_fwd, index_nonmeth_fwd = methylated_motif_occourances(motif_stripped, contig, meth_positions_fwd)
    index_meth_rev, index_nonmeth_rev = methylated_motif_occourances(motif_stripped.reverse_compliment(), contig, meth_positions_rev)

    if na_positions is not None:
        na_positions_fwd = na_positions["fwd"]
        index_meth_fwd = np.setdiff1d(index_meth_fwd, na_positions_fwd)
        index_nonmeth_fwd = np.setdiff1d(index_nonmeth_fwd, na_positions_fwd)
        na_positions_rev = na_positions["rev"]
        index_meth_rev = np.setdiff1d(index_meth_rev, na_positions_rev)
        index_nonmeth_rev = np.setdiff1d(index_nonmeth_rev, na_positions_rev)

    model = BetaBernoulliModel()
    model.update(len(index_meth_fwd) + len(index_meth_rev), len(index_nonmeth_fwd) + len(index_nonmeth_rev))
    
    if save_motif_positions:
        motif_data = {
            'index_meth_fwd': index_meth_fwd,
            'index_nonmeth_fwd': index_nonmeth_fwd,
            'index_meth_rev': index_meth_rev,
            'index_nonmeth_rev': index_nonmeth_rev
        }
        return model, motif_data
    else:
        return model


def methylated_reads_counts(
        pileup: pl.DataFrame, 
        sequence: str, 
        motif: Motif
    ) -> tuple:
    """
    Get the number of methylated and non-methylated reads for a motif.

    Parameters:
    - pileup (Pileup): The pileup to be processed. 
    - sequence (str): The sequence to be processed.
    - motif (Motif): The motif to be processed.

    Returns:
    - tuple: A tuple of two integers, the first containing the number of methylated reads, the second containing the number of non-methylated reads.
    """
    motif_index = subseq_indices(motif.string, sequence) + motif.mod_position

    pileup = pileup.filter(pl.col("position").is_in(motif_index))
    counts = pileup \
        .select(["n_mod", "n_nonmod"]) \
        .sum()
    n_mod = counts.get_column("n_mod")[0]
    n_nonmod = counts.get_column("n_nonmod")[0]
    return n_mod, n_nonmod


def motif_model_read(pileup: pl.DataFrame, contig: str, motif: Motif) -> BetaBernoulliModel:
    """
    Get the posterior for a single motif. Uses number of methylated reads as methylation count.

    Parameters:
    - pileup (Pileup): The pileup to be processed.
    - read (str): The read to be processed.
    - motif (str): The motif to be processed.

    Returns:
    - BetaBernoulliModel: The model for the candidate.
    """
    motif_stripped = motif.new_stripped_motif()
    pileup_counts = pileup.with_columns((pl.col("fraction_mod")*pl.col("Nvalid_cov")).alias("n_mod")) \
        .with_columns(pl.col("n_mod").round().cast(pl.Int32)) \
        .with_columns((pl.col("Nvalid_cov") - pl.col("n_mod")).alias("n_nonmod"))
    
    pileup_plus = pileup_counts.filter(pl.col("strand") == "+")
    pileup_minus = pileup_counts.filter(pl.col("strand") == "-")

    n_meth_plus, n_nonmeth_plus = methylated_reads_counts(pileup_plus, contig, motif_stripped)
    n_meth_minus, n_nonmeth_minus = methylated_reads_counts(pileup_minus, contig, motif_stripped.reverse_compliment())

    model = BetaBernoulliModel()
    model.update(n_meth_plus + n_meth_minus, n_nonmeth_plus + n_nonmeth_minus)
    return model


##########################################
# Motif candidate state space search
##########################################

def worker_function(
        args
    ):
    """
    Process a single subpileup for one contig and one modtype

    Parameters:
    - args (tuple): The arguments to the function: contig, modtype, subpileup
    - counter (multiprocessing.Value): The progress counter
    - lock (multiprocessing.Lock): The lock for the progress counter
    - assembly (Assembly): The assembly to be processed.
    - min_kl_divergence (float): Early stopping criteria, if max KL-divergence falls below, stops building motif.
    - padding (int): The padding to use for the motif.
    """
    (contig, modtype, subpileup, low_coverage_positions, counter, lock, assembly, 
     min_kl_divergence, padding, minimum_methylation_fraction_confident, 
     read_level_methylation, log_dir, verbose, seed) = args
    set_seed(seed=seed)
    warnings.filterwarnings("ignore")
    process_id = os.getpid()
    if log_dir is not None:
        log_file = f"{log_dir}/find-motifs.{process_id}.log"
        configure_logger(log_file, verbose=verbose)
    log.info(f"Worker {process_id} started")

    if low_coverage_positions is not None:
        low_coverage_positions = low_coverage_positions.explode("position", "strand")
        low_coverage_positions = {
            "fwd": low_coverage_positions.filter(pl.col("strand") == "+")["position"].to_numpy(),
            "rev": low_coverage_positions.filter(pl.col("strand") == "-")["position"].to_numpy()
        }
    

    try:
        result = process_subpileup(
            contig, 
            modtype, 
            subpileup, 
            assembly, 
            min_kl_divergence, 
            padding, 
            minimum_methylation_fraction_confident, 
            read_level_methylation,
            na_positions = low_coverage_positions
        )
        with lock:
            counter.value += 1
            log.debug(f"Task for ({contig}, {modtype}) complete; progress: {counter.value}")
        return result
    except Exception as e:
        log.error(f"Error processing {contig}, {modtype}: {e}")
        with lock:
            counter.value += 1
        return None


def process_subpileup(
        contig, 
        modtype, 
        contig_pileup, 
        assembly, 
        min_kl_divergence, 
        padding, 
        minimum_methylation_fraction_confident,
        read_level_methylation,
        na_positions: dict = None
    ):
    """
    Process a single subpileup for one contig and one modtype

    Parameters:
    - args (tuple): The arguments to the function: contig, modtype, subpileup
    - counter (multiprocessing.Value): The progress counter
    - lock (multiprocessing.Lock): The lock for the progress counter
    - assembly (Assembly): The assembly to be processed.
    - min_kl_divergence (float): Early stopping criteria, if max KL-divergence falls below, stops building motif.
    - padding (int): The padding to use for the motif.
    """
    log.info(f"Processing {contig} {modtype}")
    assert contig_pileup is not None, "Subpileup is None"
    assert len(contig_pileup) > 0, "Subpileup is empty"
    assert assembly is not None, "Assembly is None"
    assert min_kl_divergence >= 0, "min_kl_divergence must be greater than 0"
    assert contig_pileup.get_column("mod_type").unique().to_list() == [modtype], "subpileup modtype does not match modtype"

    contig_sequence = assembly.assembly[contig]

    motif_graph, best_candidates = find_best_candidates(
        contig_pileup, 
        contig_sequence, 
        modtype,
        minimum_methylation_fraction_confident,
        padding,
        na_positions = na_positions,
        read_level_methylation = read_level_methylation,
        min_kl = min_kl_divergence,
        max_dead_ends = 25,
        max_rounds_since_new_best = 15
    )
    identified_motifs = nxgraph_to_dataframe(motif_graph) \
        .filter(col("sequence").is_in(best_candidates))
    
    if len(identified_motifs) == 0:
        log.info("No motifs found")
        return None
    else:
        identified_motifs = identified_motifs.with_columns(
            pl.lit(contig).alias("contig"),
            pl.lit(modtype).alias("mod_type")
        ).with_columns([
            pl.col("model").map_elements(lambda x: x._alpha).alias("alpha"),
            pl.col("model").map_elements(lambda x: x._beta).alias("beta")
        ]).drop("model")
        return identified_motifs

def set_polars_env():
    os.environ["POLARS_MAX_THREADS"] = "1"



def process_sample_parallel(
        assembly, pileup, 
        low_coverage_positions = None,
        read_level_methylation = False,
        threads = 2,
        search_frame_size = 40,
        threshold_methylation_confident = 0.8,
        threshold_methylation_general = 0.7,
        threshold_valid_coverage = 5,
        minimum_kl_divergence = 0.05,
        verbose = False,
        log_dir = None,
        seed = None
    ):
    """
    Process a sample
    
    Parameters:
    - assembly (Assembly): The assembly of all contigs.
    - pileup (Pileup): The pileup to be processed.
    - max_candidate_size (int): The maximum size of the candidate motifs.
    - min_read_methylation_fraction (float): The minimum fraction of reads that must be methylated for a position to be considered methylated.
    - min_kl_divergence (float): Early stopping criteria, if max KL-divergence falls below, stops building motif.
    - min_cdf_score (float): Minimum score of 1 - cdf(cdf_position) for a motif to be considered valid.
    - cdf_position (float): The position to evaluate the cdf at.
    - min_motif_frequency (int): Used to get minimum number of sequences to evaluate motif at.
    """
    assert pileup is not None, "Pileup is None"
    assert len(pileup) > 0, "Pileup is empty" 
    assert assembly is not None, "Assembly is None"
    assert search_frame_size > 0, "search_frame_size must be greater than 0"
    assert threshold_methylation_confident >= 0 and threshold_methylation_confident <= 1, "threshold_methylation_confident must be between 0 and 1"
    assert threshold_methylation_general >= 0 and threshold_methylation_general <= 1, "threshold_methylation_general must be between 0 and 1"
    assert minimum_kl_divergence >= 0, "mininum_kl_divergence must be greater than 0"

    # Infer padding size from candidate_size
    padding = search_frame_size // 2
    
    log.debug("Partitining pileups to dict")
    pileup_dict = pileup.partition_by(["contig", "mod_type"], as_dict=True)
    low_coverage_positions_dict = low_coverage_positions.partition_by(["contig", "mod_type"], as_dict=True) if low_coverage_positions is not None else None

    def task_generator(pileup_dict, low_coverage_positions_dict):
        for (contig, modtype) in pileup_dict:
            subpileup = pileup_dict[(contig, modtype)]
            
            if low_coverage_positions is not None:
                if (contig, modtype) in low_coverage_positions_dict.keys():
                    low_coverage_positions_filtered = low_coverage_positions_dict[(contig, modtype)]
                else:
                    low_coverage_positions_filtered = None
            else:
                low_coverage_positions_filtered = None

            yield (contig, modtype, subpileup, low_coverage_positions_filtered, counter, lock, assembly, 
                       minimum_kl_divergence, search_frame_size // 2, threshold_methylation_confident, 
                       read_level_methylation, log_dir, verbose, seed)
                
    # Create a progress manager
    manager = multiprocessing.Manager()
    counter = manager.Value('i', 0)
    lock = manager.Lock()

    # Create a pool of workers
    pool = get_context("spawn").Pool(processes=threads, initializer=set_polars_env)

    # Create a process for the progress bar
    log.debug("Counting number of tasks")
    total_tasks = len(pileup.select("contig", "mod_type").unique())

    log.debug("Starting progress bar")
    progress_bar_process = multiprocessing.Process(target=update_progress_bar, args=(counter, total_tasks, True))
    progress_bar_process.start()

    # Put them workers to work
    log.debug("Starting workers")
    chunksize = max(min(100, total_tasks // threads), 1)
    results = pool.imap(worker_function, task_generator(pileup_dict, low_coverage_positions_dict), chunksize=chunksize)
    results = [result for result in results if result is not None]
    log.debug("Joining results")

    log.debug("Closing pool")
    pool.close()
    pool.join()

    log.debug("Joining progress bar")
    progress_bar_process.join()

    log.debug("Creating final dataframe")
    if len(results) == 0:
        log.debug("No motifs found")
        return None
    motifs = pl.concat(results, rechunk=True, parallel=False)

    model_col = []
    for a, b in zip(motifs.get_column("alpha").to_list(), motifs.get_column("beta").to_list()):
        model_col.append(BetaBernoulliModel(a, b))


    motifs = motifs.with_columns([
            pl.Series(model_col).alias("model"),
            pl.col("sequence").map_elements(lambda motif: motif[count_periods_at_start(motif):len(motif)-count_periods_at_end(motif)]).alias("motif"),
            pl.col("sequence").map_elements(lambda motif: padding - count_periods_at_start(motif)).alias("mod_position")
        ]).drop(["alpha", "beta"])

    return motifs




#########################################################################
# Motif candidate state space 

def find_best_candidates(        
        contig_pileup: pl.DataFrame, 
        contig_sequence: DNAsequence, 
        mod_type: str,
        minimum_methylation_fraction_confident: float,
        padding: int,
        na_positions = None,
        read_level_methylation: bool = False,
        min_kl: float = 0.2, 
        max_dead_ends: int = 25, 
        max_rounds_since_new_best: int = 15,
        score_threshold: float = 0.2,
        remaining_sequences_threshold: float = 0.01
    ) -> tuple[MotifTree, list[Motif]]:
    """
    Find the best motif candidates in a sequence.

    Parameters:
    - contig_pileup (Pileup): The pileup to be processed.
    - contig_sequence (DNAsequence): The sequence to be processed.
    - mod_type (str): The modtype to be processed.
    - minimum_methylation_fraction_confident (float): The minimum fraction of reads that must be methylated for a position to be considered confidently methylated and used in search.
    - padding (int): The padding to use for the motif.
    - min_kl (float): The minimum KL-divergence for a motif to be considered valid.
    - max_dead_ends (int): The maximum number of low scoring candidates before stopping the search.
    - max_rounds_since_new_best (int): The maximum number of rounds since a new best candidate was found before stopping the search.
    - score_threshold (float): The minimum score for a candidate to be considered valid.
    - remaining_sequences_threshold (float): The minimum fraction of sequences remaining before stopping the search.
    """
    subpileup_confident = contig_pileup.filter(pl.col("fraction_mod") >= minimum_methylation_fraction_confident)

    # Extract the sequences for confidently methylated positions
    index_plus = subpileup_confident.filter(pl.col("strand") == "+").get_column("position").to_list()
    index_minus = subpileup_confident.filter(pl.col("strand") == "-").get_column("position").to_list()
    
    # Sample sequences in contig to get background for KL-divergence
    contig_sequences_sample = contig_sequence.sample_n_subsequences(
        padding * 2 + 1, 10000
    )
    contig_pssm = contig_sequences_sample.pssm()

    methylation_sequences_string = []
    if len(index_plus) >= 1:
        methylation_sequences_string += contig_sequence.sample_at_indices(index_plus, padding).sequences
    if len(index_minus) >= 1:
        methylation_sequences_string += contig_sequence.sample_at_indices(index_minus, padding).reverse_compliment().sequences
    if len(methylation_sequences_string) == 0:
        log.info("No methylation sequences found")
        return None
    methylation_sequences = EqualLengthDNASet(methylation_sequences_string).convert_to_DNAarray()

    total_sequences = methylation_sequences.shape[0]
    
    root_motif = Motif("." * padding + MOD_TYPE_TO_CANONICAL[mod_type] + "." * padding, padding)
    methylation_sequences_clone = methylation_sequences.copy()
    best_candidates = []
    continue_search = True
    dead_ends = 0
    motif_graph = None

    while continue_search:
        if dead_ends >= max_dead_ends:
            log.debug("Stopping search, too many low scoring candidates")
            break
        # Find the initial guess within the tree
        searcher = MotifSearcher(
            root_motif, 
            contig_sequence, 
            contig_pssm,
            contig_pileup, 
            methylation_sequences_clone, 
            padding,
            read_level_methylation = read_level_methylation,
            na_positions = na_positions,
            motif_graph = motif_graph,
            min_kl = min_kl,
            max_rounds_since_new_best = max_rounds_since_new_best
        )
        motif_graph, naive_guess = searcher.run()

        # If there is no naive guess, stop the search
        if naive_guess == root_motif:
            log.debug("No naive guess found, stopping search")
            break

        # Remove new candidate from methylation sequences
        seq_before = methylation_sequences.shape[0]
        methylation_sequences_clone = methylation_sequences_clone.filter_sequence_matches(naive_guess.one_hot(), keep_matches = False)
        if methylation_sequences_clone is None:
            log.debug("No more sequences left")
            break

        # Check if we should continue the search
        seq_remaining = methylation_sequences_clone.shape[0]
        seq_remaining_percent = seq_remaining/total_sequences

        if motif_graph.nodes[naive_guess]["score"] < score_threshold:
            dead_ends += 1
            log.debug(f"Candidate has low score, {naive_guess}. {dead_ends} of {max_dead_ends} before temination")
            continue

        log.info(f"Keeping {naive_guess}, represented in {seq_before-seq_remaining} seqs. model: {motif_graph.nodes[naive_guess]['model']}. ({100*seq_remaining_percent:.1f} % of sequences remaining)")

        best_candidates.append(naive_guess)
        
        if (seq_remaining/total_sequences) < remaining_sequences_threshold:
            log.debug("Stopping search, too few sequences remaining")
            break
        log.debug("Continuing search")
    return motif_graph, best_candidates



class MotifSearcher:
    """
    Class to perform motif search for identifying enriched motifs in methylated sequences.
    """

    def __init__(
        self,
        root_motif: Motif,
        contig_sequence: DNAsequence,
        contig_pssm: DNAarray,
        contig_pileup: pl.DataFrame,
        methylation_sequences: DNAarray,
        padding: int,
        read_level_methylation: bool = False,
        na_positions = None,
        motif_graph: Optional[MotifTree] = None,
        min_kl: float = 0.1,
        freq_threshold: float = 0.25,
        max_rounds_since_new_best: int = 10,
        max_motif_length: int = 18
    ):
        """
        Initialize the MotifSearcher class.

        Parameters:
            root_motif (Motif): The initial motif to start the search from.
            contig_sequence: The contig sequence object with necessary methods.
            contig_pssm: The position-specific scoring matrix for the contig background.
            contig_pileup: The pileup data associated with the contig.
            methylation_sequences: The methylation sequences to be analyzed.
            padding (int): The padding size for sequence sampling.
            read_level_methylation (bool): Flag to use read-level methylation.
            na_positions (Optional[List[int]]): Positions with 'N/A' or missing data.
            motif_graph (Optional[MotifTree]): An optional pre-initialized motif graph.
            min_kl (float): Minimum Kullback-Leibler divergence threshold.
            freq_threshold (float): Frequency threshold for methylation.
            max_rounds_since_new_best (int): Max iterations without improvement.
            max_motif_length (int): Maximum allowed motif length.
        """
        # Input validation
        if contig_pileup.is_empty():
            raise ValueError("contig_pileup cannot be empty.")
        if padding < 0:
            raise ValueError("padding must be non-negative.")

        self.root_motif = root_motif
        self.contig_sequence = contig_sequence
        self.contig_pssm = contig_pssm
        self.contig_pileup = contig_pileup
        self.methylation_sequences = methylation_sequences
        self.padding = padding
        self.read_level_methylation = read_level_methylation
        self.na_positions = na_positions
        self.motif_graph = motif_graph or MotifTree()
        self.min_kl = min_kl
        self.freq_threshold = freq_threshold
        self.max_rounds_since_new_best = max_rounds_since_new_best
        self.max_motif_length = max_motif_length

    def _priority_function(self, next_model, root_model) -> float:
        """
        Calculate the priority for a motif based on the change in model parameters.

        Parameters:
            next_model: The model of the next motif.
            root_model: The model of the root motif.

        Returns:
            float: The calculated priority.
        """
        try:
            d_alpha = 1 - (next_model._alpha / root_model._alpha)
        except ZeroDivisionError:
            d_alpha = 1
        try:
            d_beta = next_model._beta / root_model._beta
        except ZeroDivisionError:
            d_beta = 1
        priority = d_alpha * d_beta
        return priority

    def _scoring_function(self, next_model, current_model) -> float:
        """
        Calculate the score for a motif based on its model.

        Parameters:
            next_model: The model of the next motif.
            current_model: The model of the current motif.

        Returns:
            float: The calculated score.
        """
        mean_diff = next_model.mean() - current_model.mean()
        try:
            score = (next_model.mean() * -np.log10(next_model.standard_deviation())) * mean_diff
        except (ValueError, ZeroDivisionError):
            score = 0
        return score

    def _motif_child_nodes_kl_dist_max(
        self,
        motif: Motif,
        meth_pssm: np.ndarray,
        contig_pssm: np.ndarray
    ) -> Generator[Motif, None, None]:
        """
        Generate child motifs based on maximum KL divergence.

        Parameters:
            motif (Motif): The current motif.
            meth_pssm (np.ndarray): Position-specific scoring matrix for methylation.
            contig_pssm (np.ndarray): Position-specific scoring matrix for the contig.

        Yields:
            Motif: The next motif in the search.
        """
        kl_divergence = entropy(meth_pssm, contig_pssm)
        split_motif = motif.split()

        # Positions to evaluate (positions with '.')
        evaluated_positions = np.array([i for i, base in enumerate(split_motif) if base == "."])
        if evaluated_positions.size == 0:
            return

        kl_divergence_masked = kl_divergence.copy()
        kl_divergence_masked[~np.isin(np.arange(len(split_motif)), evaluated_positions)] = 0

        max_kl = np.max(kl_divergence_masked)
        if max_kl < self.min_kl:
            return

        pos = int(np.argmax(kl_divergence_masked))

        # Methylation frequency must be above contig frequency
        index_meth_freq_higher = meth_pssm[:, pos] > contig_pssm[:, pos]

        # Methylation frequency must be above a threshold
        index_meth_freq_above_thresh = meth_pssm[:, pos] > self.freq_threshold

        # Combine the two filters
        index_position_filter = np.logical_and(index_meth_freq_higher, index_meth_freq_above_thresh)
        bases_indices = np.argwhere(index_position_filter).reshape(-1)
        bases_filtered = [BASES[int(i)] for i in bases_indices]

        if not bases_filtered:
            return

        # All combinations of the bases
        max_combination_length = min(len(bases_filtered), 4)
        for i in range(1, max_combination_length + 1):
            for base_tuple in itertools.combinations(bases_filtered, i):
                if len(base_tuple) > 1:
                    base_str = "[" + "".join(base_tuple) + "]"
                else:
                    base_str = base_tuple[0]
                new_motif_sequence = split_motif.copy()
                new_motif_sequence[pos] = base_str
                new_motif_str = "".join(new_motif_sequence)
                yield Motif(new_motif_str, motif.mod_position)

    def run(self) -> tuple[MotifTree, Motif]:
        """
        Execute the motif search algorithm.

        Returns:
            Tuple[MotifTree, Motif]: The graph of motifs explored and the best motif found.
        """
        # Initialize variables
        best_guess = self.root_motif
        if self.read_level_methylation:
            root_model = motif_model_read(
                self.contig_pileup, self.contig_sequence.sequence, self.root_motif
            )
        else:
            root_model = motif_model_contig(
                self.contig_pileup,
                self.contig_sequence.sequence,
                self.root_motif,
                na_positions=self.na_positions
            )
        best_score = self._scoring_function(root_model, root_model)
        rounds_since_new_best = 0
        visited_nodes: set[Motif] = set()
        
        # Initialize the search tree
        self.motif_graph.add_node(
            self.root_motif,
            model=root_model,
            motif=self.root_motif,
            visited=False,
            score=best_score
        )

        # Initialize priority queue
        priority_queue: list[tuple[float, Motif]] = []
        hq.heappush(priority_queue, (0, self.root_motif))

        # Search loop
        while priority_queue:
            # Get the current best candidate
            _, current_motif = hq.heappop(priority_queue)
            if current_motif in visited_nodes:
                continue
            if len(current_motif.strip()) > self.max_motif_length:
                log.debug(
                    f"{current_motif.string}, Skipping due to length > {self.max_motif_length}"
                )
                continue

            current_model = self.motif_graph.nodes[current_motif]["model"]
            visited_nodes.add(current_motif)
            log.debug(
                f"{current_motif.string} | Model: {current_model} | "
                f"Score: {self.motif_graph.nodes[current_motif]['score']:.2f} | "
                f"Queue size: {len(priority_queue)}"
            )
            self.motif_graph.nodes[current_motif]["visited"] = True
            rounds_since_new_best += 1

            active_methylation_sequences = self.methylation_sequences.copy().filter_sequence_matches(
                current_motif.one_hot(), keep_matches=True
            )
            if active_methylation_sequences is None:
                log.debug("No more sequences left after filtering")
                continue

            # Generate and evaluate neighbor motifs
            neighbors = list(self._motif_child_nodes_kl_dist_max(
                current_motif,
                active_methylation_sequences.pssm(),
                self.contig_pssm
            ))

            # Add neighbors to graph
            for next_motif in neighbors:
                if next_motif in self.motif_graph.nodes:
                    # Add only edge if motif already visited
                    next_model = self.motif_graph.nodes[next_motif]["model"]
                    score = self.motif_graph.nodes[next_motif]["score"]
                    self.motif_graph.add_edge(current_motif, next_motif)
                else:
                    if self.read_level_methylation:
                        next_model = motif_model_read(
                            self.contig_pileup,
                            self.contig_sequence.sequence,
                            next_motif
                        )
                    else:
                        next_model = motif_model_contig(
                            self.contig_pileup,
                            self.contig_sequence.sequence,
                            next_motif,
                            na_positions=self.na_positions
                        )

                    # Add neighbor to graph
                    self.motif_graph.add_node(
                        next_motif,
                        model=next_model,
                        motif=next_motif,
                        visited=False
                    )
                    self.motif_graph.add_edge(current_motif, next_motif)

                    score = self._scoring_function(next_model, current_model)
                    self.motif_graph.nodes[next_motif]["score"] = score

                # Add neighbor to priority queue if not visited
                if next_motif not in visited_nodes:
                    priority = self._priority_function(next_model, current_model)
                    hq.heappush(priority_queue, (priority, next_motif))

                if score > best_score:
                    best_score = score
                    best_guess = next_motif
                    rounds_since_new_best = 0

            # Stopping criteria
            if rounds_since_new_best >= self.max_rounds_since_new_best:
                break

        return self.motif_graph, best_guess




def count_periods_at_start(string: str) -> int:
    """
    Count the number of periods at the start of a string

    Parameters:
    - string (str): The string to be processed.

    Returns:
    - int: The number of periods at the start of the string.

    Example:
    >>> count_periods_at_start("...ACGT")
    3
    """
    count = 0
    for char in string:
        if char == '.':
            count += 1
        else:
            break
    return count
def count_periods_at_end(string):
    string = string[::-1]
    count = 0
    for char in string:
        if char == '.':
            count += 1
        else:
            break
    return count
def nxgraph_to_dataframe(graph):
    return pl.DataFrame({
        "sequence":[i for i in graph.nodes],
        "model":[graph.nodes[i]["model"] for i in graph.nodes],
        "score":[graph.nodes[i]["score"] for i in graph.nodes]
    }).sort("score", descending=True)





import multiprocessing
import progressbar
import time


def update_progress_bar(progress, total_tasks, print_progress, timeout=30000):
    """
    Update progress bar for parallel processing

    Parameters:
    - progress (multiprocessing.Value): The progress counter
    - total_tasks (int): The total number of tasks
    - timeout (int): The timeout in seconds
    """
    last_update_time = time.time()
    with progressbar.ProgressBar(max_value=total_tasks) as bar:
        while True:
            current_time = time.time()
            if progress.value >= total_tasks:
                break
            if current_time - last_update_time > timeout:
                print("Timeout reached. Exiting progress bar update.")
                break
            if print_progress:
                bar.update(progress.value)
            time.sleep(0.5)



import os
import subprocess
import sys
import shutil
import snakemake

def snakemake_create_environments(args):

    thisdir = os.path.abspath(os.path.dirname(__file__))
    cwd = os.getcwd()

    snakefile_path = os.path.join(thisdir, "setup.smk")
    if os.path.exists(snakefile_path):
        snakefile = snakefile_path
        print(snakefile_path)
    else:
        msg = 'Error: cannot find Snakefile at the following location:\n'
        msg += '{}\n'.format(snakefile_path)
        sys.stderr.write(msg)
        sys.exit(-1)

    dependency_dir = os.path.join(args.dependency_dir, "ML_dependencies")


    command = [
        "snakemake",
        "--snakefile", snakefile,
        "--config", f"DEPENDENCYDIR={dependency_dir}",
        "--use-conda",
        "--conda-prefix", os.path.join(dependency_dir, "ML_envs"),
        "--conda-create-envs-only",
        "--cores", "1",
        "all"
        ]
        
    try:
        result = subprocess.run(command, cwd=cwd, check=True)
        if result.returncode == 0:
            print("Conda environments were successfully created.")
    except subprocess.CalledProcessError as e:
        print("Creation of conda environments failed.", e)
        sys.exit(1)


def get_models(args):

    thisdir = os.path.abspath(os.path.dirname(__file__))
    cwd = os.getcwd()

    snakefile_path = os.path.join(thisdir, "setup.smk")
    if os.path.exists(snakefile_path):
        snakefile = snakefile_path
        print(snakefile_path)
    else:
        msg = 'Error: cannot find Snakefile at the following location:\n'
        msg += '{}\n'.format(snakefile_path)
        sys.stderr.write(msg)
        sys.exit(-1)

    dependency_dir = os.path.join(args.dependency_dir, "ML_dependencies")


    command = [
        "snakemake",
        "--snakefile", snakefile,
        "--config", f"DEPENDENCYDIR={dependency_dir}",
        "--use-conda",
        "--conda-prefix", os.path.join(dependency_dir, "ML_envs"),
        "--cores", "1",
        "make_REbase_db"
        ]
        
    try:
        result = subprocess.run(command, cwd=cwd, check=True)
        if result.returncode == 0:
            print("pfam models and REbase database were successfully retreived.")
    except subprocess.CalledProcessError as e:
        print("pfam models and REbase database could not be retreived.", e)
        sys.exit(1)



def defensefinder_update(args):

    thisdir = os.path.abspath(os.path.dirname(__file__))
    cwd = os.getcwd()

    snakefile_path = os.path.join(thisdir, "setup.smk")
    if os.path.exists(snakefile_path):
        snakefile = snakefile_path
        print(snakefile_path)
    else:
        msg = 'Error: cannot find Snakefile at the following location:\n'
        msg += '{}\n'.format(snakefile_path)
        sys.stderr.write(msg)
        sys.exit(-1)

    dependency_dir = os.path.join(args.dependency_dir, "ML_dependencies")


    command = [
        "snakemake",
        "--snakefile", snakefile,
        "--config", f"DEPENDENCYDIR={dependency_dir}",
        "--use-conda",
        "--conda-prefix", os.path.join(dependency_dir, "ML_envs"),
        "--cores", "1",
        "defense_finder_update"
        ]
        
    try:
        result = subprocess.run(command, cwd=cwd, check=True)
        if result.returncode == 0:
            print("Defensefinder models updated succesfully")
    except subprocess.CalledProcessError as e:
        print("Update of Defensefinder models failed.", e)
        sys.exit(1)


def check_installation_MTase_linker(args):

    thisdir = os.path.abspath(os.path.dirname(__file__))
    cwd = os.getcwd()

    dependency_dir = os.path.join(args.dependency_dir, "ML_dependencies")

    command = [
        "nanomotif", 
        "MTase-linker", 
        "run", 
        "--threads", "1", 
        "--assembly", os.path.join(thisdir, "..", "datasets", "e_coli_assembly.polished.fasta"), 
        "--contig_bin", os.path.join(thisdir, "..", "datasets", "e_coli_contig-bin.tsv"), 
        "--bin_motifs", os.path.join(thisdir, "..", "datasets", "e_coli_bin-motifs.tsv"), 
        "--out", os.path.join(dependency_dir, "check_installation"),
        "-d", dependency_dir, 
        "--forceall", "True"
        ]

    print("Testing installation of MTase-linker.")

    try:
        result = subprocess.run(command, cwd=cwd, check=True)
        if result.returncode == 0:
            print("Installation of MTase-linker was successful.")
    except subprocess.CalledProcessError as e:
        print("Installation of MTase-linker failed.", e)
        sys.exit(1)


import subprocess
import os
import sys
import glob

def run_MTase_linker(args):
    
    thisdir = os.path.abspath(os.path.dirname(__file__))
    cwd = os.getcwd()

    snakefile_path = os.path.join(thisdir,"MTase_linker.smk")
    if os.path.exists(snakefile_path):
        snakefile = snakefile_path
        print(snakefile_path)
    else:
        msg = 'Error: cannot find Snakefile at the following location:\n'
        msg += '{}\n'.format(snakefile_path)
        sys.stderr.write(msg)
        sys.exit(1)

    # Bins directory
    a1 = os.path.join(cwd, args.assembly)
    a2 = os.path.join(args.assembly)
    if os.path.exists(a1):
        assembly_path = a1
    elif os.path.exists(a2):
        assembly_path = a2
    else:
        msg = 'Error: cannot find assembly files in paths {} or {} '.format(a1,a2)
        sys.stderr.write(msg)
        sys.exit(1)


    # contig_bin.tsv
    b1 = os.path.join(cwd, args.contig_bin)
    b2 = os.path.join(args.contig_bin)
    if os.path.exists(b1) and not os.path.isdir(b1) and os.path.splitext(b1)[1] == '.tsv':
        contig_bin = b1
    elif os.path.exists(b2) and not os.path.isdir(b2) and os.path.splitext(b2)[1] == '.tsv':
        contig_bin = b2
    elif os.path.exists(b1) and not os.path.isdir(b1):
        msg = 'Error: contig_bin file {} does not have .tsv. extension'.format(b1)
        sys.stderr.write(msg)
        sys.exit(1)
    elif os.path.exists(b2) and not os.path.isdir(b2):
        msg = 'Error: contig_bin file {} does not have .tsv. extension'.format(b2)
        sys.stderr.write(msg)
        sys.exit(1)
    else:
        msg = 'Error: cannot find bins directory in paths {} or {} '.format(b1,b2)
        sys.stderr.write(msg)
        sys.exit(1)


    c1 = os.path.join(cwd, args.dependency_dir)
    c2 = os.path.join(args.dependency_dir)
    if os.path.exists(c1) and os.path.isdir(c1):
        dependency_dir = c1
    elif os.path.exists(c2) and os.path.isdir(c2):
        dependency_dir = c2
    else:
        msg = 'Dependency dir not found at {} or {}, run MTase-linker install'.format(c1, c2)
        sys.stderr.write(msg)
        sys.exit(1)
    
    if os.path.exists(os.path.join(c1, "df_models", "defense-finder_update.done")):
        print("Defensefinder models found!")
    elif os.path.exists(os.path.join(c2, "df_models", "defense-finder_update.done")):
            print("Defensefinder models found!")
    else:
        msg = 'Defensefinder models not found at {} or {}, run MTase-linker install'.format(c1, c2)
        sys.stderr.write(msg)
        sys.exit(1)

    if os.path.exists(os.path.join(c1, "ML_envs")):
        print("Environments found!")
    elif os.path.exists(os.path.join(c2, "ML_envs")):
            print("Environments found!")
    else:
        msg = 'Environments not found at {} or {}, run MTase-linker install'.format(c1, c2)
        sys.stderr.write(msg)
        sys.exit(1)

    if os.path.exists(os.path.join(c1, "REbase_RecSeq_all")):
        print("REbase database found!")
    elif os.path.exists(os.path.join(c2, "REbase_RecSeq_all")):
            print("REbase database found!")
    else:
        msg = 'REbase database not found at {} or {}, run MTase-linker install'.format(c1, c2)
        sys.stderr.write(msg)
        sys.exit(1)

    if os.path.exists(os.path.join(c1, "PFAM_MTase_profiles.hmm")):
        print("PFAM models found!")
    elif os.path.exists(os.path.join(c2, "PFAM_MTase_profiles.hmm")):
            print("PFAM models found!")
    else:
        msg = 'PFAM models not found at {} or {}, run MTase-linker install'.format(c1, c2)
        sys.stderr.write(msg)
        sys.exit(1)

    command = [
        "snakemake",
        "--snakefile", snakefile,
        "--cores", str(args.threads),
        "--config",
        f"THREADS={args.threads}",
        f"ASSEMBLY={assembly_path}",
        f"CONTIG_BIN={contig_bin}",
        f"OUTPUTDIRECTORY={args.out}",
        f"DEPENDENCY_PATH={dependency_dir}",
        f"IDENTITY={args.identity}",
        f"QCOVS={args.qcovs}",
        f"NANOMOTIF={args.bin_motifs}",
        "--use-conda",
        "--conda-prefix", os.path.join(dependency_dir, "ML_envs")
        
        ]
    
    if args.forceall:
        command.append("--forceall")
    if args.dryrun:
        command.append("--dryrun")
        
    try:
        result = subprocess.run(command, cwd=cwd, check=True)
        if result.returncode == 0:
            print("MTase-Linker done!")
    except subprocess.CalledProcessError as e:
        print("MTase-linker failed with error:", e)
        sys.exit(1)

    # workflow = None
    # workflow = {"THREADS": args.threads,
    #             "ASSEMBLY": assembly_path,
    #             "CONTIG_BIN": contig_bin,
    #             "OUTPUTDIRECTORY": args.out,
    #             "DEPENDENCY_PATH": dependency_dir,
    #             "IDENTITY": args.identity,
    #             "QCOVS": args.qcovs,
    #             "NANOMOTIF": args.bin_motifs}


    # status = snakemake.snakemake(snakefile, 
    #                             config=workflow,
    #                             targets = ["all"], 
    #                             use_conda = True, 
    #                             forceall=args.forceall, 
    #                             cores = args.threads,
    #                             dryrun=args.dryrun, 
    #                             workdir = cwd,
    #                             conda_prefix = os.path.join(dependency_dir, "ML_envs"))

    # if status:
    #     print("MTase-Linker done!")
    # else:
    #     print("MTase-linker failed.")




import pytest
from nanomotif.binnary import data_processing
from .conftest import MockArgs
import polars as pl

def test_feature_with_loaded_data(loaded_data):
    # Access the loaded data directly if returned as a dictionary
    motifs_scored = loaded_data["motifs_scored"]
    bin_motifs = loaded_data["bin_motifs"]
    contig_bins = loaded_data["contig_bins"]

    # Now you can use the data in your test assertions
    assert motifs_scored is not None

    # bin motifs
    assert bin_motifs is not None
    assert set(bin_motifs["bin"].unique()) == {"b1", "b2", "b3"}

    # contig bins
    assert contig_bins is not None
    assert set(contig_bins["bin"].unique()) == {"b1", "b2", "b3", "b4"}

    contig_set = {f"contig_{i}" for i in range(1, 14)} | {"contig_16"}
    assert set(contig_bins["contig"].unique()) == contig_set



def test_prepare_bin_consensus(loaded_data):
    """
    GIVEN loaded_data
    WHEN prepare_bin_consensus is called
    THEN assert that the output contains only the expected columns
    """
    args = MockArgs()
    
    bin_motif_binary = data_processing.prepare_bin_consensus(loaded_data["bin_motifs"], args)
    
    assert bin_motif_binary is not None
    assert bin_motif_binary.columns == ["bin", "motif_mod", "mean_methylation", "methylation_binary"]
    
    assert bin_motif_binary.filter((pl.col("bin") == "b3") & (pl.col("motif_mod") == "m6_a_1")).select("methylation_binary").item() == 1
    
    # Assert that there are 4 motifs in bin 1
    assert bin_motif_binary.filter(pl.col("bin") == "b3").height == 4
    assert set(bin_motif_binary.filter(pl.col("bin") == "b3").to_pandas()["motif_mod"].to_list()) == {"m1_a_1", "m2_a_1", "m3_a_1", "m6_a_1"}
    
    # Assert m7_a is filtered because of too few observations
    assert bin_motif_binary.filter(pl.col("motif_mod") == "m7_a_1").is_empty()
    


def test_add(loaded_data):
    # Access the loaded data directly if returned as a dictionary
    motifs_scored = loaded_data["motifs_scored"]
    bin_motifs = loaded_data["bin_motifs"]
    contig_bins = loaded_data["contig_bins"]

    # Step 1 create bin_motif_binary
    args = MockArgs()
    bin_motif_binary = data_processing.prepare_bin_consensus(bin_motifs, args)

    # Step 2: create motifs_scored_in_bins
    motifs_scored_in_bins = data_processing.add_bin(
        motifs_scored,
        contig_bins
    )
    print(motifs_scored_in_bins)
    # motifs_scored_in_bins
    assert sorted(bin_motif_binary.select("motif_mod").unique()["motif_mod"]) == sorted(
        motifs_scored_in_bins.select("motif_mod").unique()["motif_mod"]
    )
    
    # Assert that the number of columns is 12
    assert set(motifs_scored_in_bins.columns) == set(['contig', 'median', 'N_motif_obs', 'motif_mod', 'bin']) # number of columns
    # Assert contig 1 belongs to bin 1
    assert motifs_scored_in_bins.filter(pl.col("contig") == "contig_1").select("bin").unique()["bin"].item() == "b1"
    

   


def test_impute_contig_methylation_within_bin(loaded_data, motifs_scored_in_bins_and_bin_motifs):
    """
    GIVEN loaded_data and motifs_scored_in_bins_and_bin_motifs
    WHEN construct_bin_motifs_from_motifs_scored_in_bins is called
    THEN assert that the output contains only the expected columns
    """
    motifs_scored_in_bins = motifs_scored_in_bins_and_bin_motifs["motifs_scored_in_bins"]
    
    args = MockArgs()
    print(motifs_scored_in_bins)
    imputed_binned_contig_methylation = data_processing.impute_contig_methylation_within_bin(
        motifs_scored_in_bins
    )
    print(imputed_binned_contig_methylation)
    assert imputed_binned_contig_methylation is not None
    assert set(imputed_binned_contig_methylation.columns) == set(['contig', 'bin', 'motif_mod', 'mean_bin_median', 'median' ])
    assert None not in imputed_binned_contig_methylation.to_pandas()["median"] 
    
def test_impute_contig_methylation_within_bin2():
    # Sample input DataFrame
    contig_methylation = pl.DataFrame({
        "contig": ["contig_1", "contig_1", "contig_1", "contig_2", "contig_3"],
        "bin": ["bin1", "bin1", "bin1", "bin1", "bin2"],
        "motif_mod": ["mod1", "mod2", "mod3", "mod3", "mod2"],
        "median": [0.5, 0.0, 0.9, 0.5, 0.9],
        "N_motif_obs": [10, 5, 15, 20, 25]
    })

    # Expected output DataFrame after imputation
    expected_output = pl.DataFrame({
        "contig": ["contig_1", "contig_1", "contig_1", "contig_2", "contig_2", "contig_2", "contig_3"],
        "bin": ["bin1", "bin1", "bin1", "bin1", "bin1", "bin1", "bin2"],
        "motif_mod": ["mod1", "mod2","mod3", "mod1","mod2", "mod3","mod2"],
        "median": [0.5, 0.0, 0.9, 0.5, 0.0, 0.5, 0.9]
    })

    # Since args is not used in the function (commented out), we can pass None
    args = None

    # Call the function to test
    output = data_processing.impute_contig_methylation_within_bin(contig_methylation)


    # Select and sort columns for comparison
    output = output.select(expected_output.columns).sort(["bin", "contig", "motif_mod"]).drop("mean_bin_median")
    expected_output = expected_output.sort(["bin", "contig", "motif_mod"])
    print(output)
    print(expected_output)

    # Assert that the output matches the expected output
    assert output.equals(expected_output, null_equal=True), "The imputed DataFrame does not match the expected output."
    
def test_impute_unbinned_contigs():
    import numpy as np
    np.random.seed(42)
    # Mock input data
    contig_methylation = pl.DataFrame({
        "bin": ["unbinned", "binned", "unbinned", "binned"],
        "contig": ["contig_1", "contig_2", "contig_3", "contig_4"],
        "motif_mod": ["mod1", "mod2", "mod1", "mod3"],
        "median": [0.0, 0.3, 0.95, 0.5]
    })

    # Define the expected behavior of the pseudo-random number generator
    result = data_processing.impute_unbinned_contigs(contig_methylation)
    print(result)

    # Validate the resulting DataFrame structure
    assert "median" in result.columns
    assert len(result.get_column("contig")) == 6
    assert set(result.get_column("motif_mod").unique()) == set(["mod1", "mod2", "mod3"])
    assert None not in result.to_pandas()["median"]
    # Expected values for the "median" column (based on your output)
    expected_median = [0.0, 0.142607, 0.109799, 0.95, 0.023403, 0.023399]

    # Extract actual "median" values from the DataFrame
    actual_median = result.get_column("median").to_list()

    # Assert the two lists are equal with some tolerance for floating-point precision
    assert np.allclose(actual_median, expected_median, atol=1e-6), \
        f"Median values do not match: {actual_median} != {expected_median}"

    




"""
Models
"""
import numpy as np
from scipy.stats import beta

DEFAULT_PRIOR_BETA = 1
DEFAULT_PRIOR_ALPHA = 0

class BetaBernoulliModel():
    def __init__(self, alpha = DEFAULT_PRIOR_ALPHA, beta = DEFAULT_PRIOR_BETA):
        self.alpha = alpha
        self.beta = beta
        self._alpha = alpha
        self._beta = beta

    def update(self, n_positives, n_negatives):
        self._alpha += n_positives
        self._beta += n_negatives

    def reset(self):
        self._alpha = self.alpha
        self._beta = self.beta

    def sample(self):
        return np.random.beta(self._alpha, self._beta)

    def mean(self):
        return self._alpha / (self._alpha + self._beta)

    def variance(self):
        a = self._alpha
        b = self._beta
        return (a * b) / ((a + b)**2 * (a + b + 1))
    
    def standard_deviation(self):
        return np.sqrt(self.variance())
    
    def cdf(self, x):
        return beta.cdf(x, self._alpha, self._beta)

    def __str__(self):
        return f'BetaBernoulliModel(alpha={self._alpha}, beta={self._beta})'

    def __repr__(self):
        return str(self)





# pyright: reportUndefinedVariable=false
#%% load libraries
import pandas as pd
import os
from utillities import mod_predictions_hmm

#%% Defining pfam hit table ath
pfam_hit_tsv_path = snakemake.input[0]

if os.path.getsize(pfam_hit_tsv_path) > 0: 
    
    # Importing pfam hit table
    pfam_hit_df = pd.read_csv(pfam_hit_tsv_path, sep = '\t', header = None)
    pfam_hit_df = pfam_hit_df.loc[:, 0:3]
    pfam_hit_df.columns = ["gene_id", "acc", "HMM_name", "HMM_acc"]
#%%
    #Adding modtype prediction to pfam hits
    pfam_hit_mod_df = mod_predictions_hmm(pfam_hit_df, 'HMM_acc')

    def process_mod_type(group):

        mod_types_concatenated = ', '.join(group['mod_type'])
    
        mod_types_condensed = ', '.join(sorted(set(mod_types_concatenated.split(', '))))
    
    # Check if both 'm' and 'a' are present, if so, choose 'm' only
        if 'm' in mod_types_condensed and 'ac' in mod_types_condensed:
            mod_types_condensed = mod_types_condensed.replace('ac, ', '').replace(', ac', '').replace('ac', '')
    

        return mod_types_condensed

    protein_acc_mod_df = pfam_hit_mod_df.groupby('gene_id').apply(process_mod_type).reset_index()
    protein_acc_mod_df.columns = ['gene_id', 'mod_type']
#%%
    protein_acc_mod_df.to_csv(snakemake.output[0] , sep='\t', index=False)

else:
    open(snakemake.output[0], 'w').close()




import polars as pl
import pandas as pd
import nanomotif.binnary.utils as ut
from pymethylation_utils.utils import run_epimetheus
from .conftest import MockArgs
import os
from pathlib import Path

def test_split_bin_contig():
    # Input DataFrame
    input_data = pl.DataFrame({
        "contig_bin": ["bin1", "bin1", "bin_c1"],
        "bin_compare": ["bin1_contig_23", "bin1_contig_40", "bin_c1_contig_23"]
    })

    # Expected output DataFrame
    expected_output = pd.DataFrame({
        "contig_bin": ["bin1", "bin1", "bin_c1"],
        "bin_compare": ["bin1_contig_23", "bin1_contig_40", "bin_c1_contig_23"],
        "contig": ["contig_23", "contig_40", "contig_23"]
    })

    # Run the function
    output = ut.split_bin_contig(input_data)
    output = output.to_pandas()
    print(output)
    
    # Verify the output
    pd.testing.assert_frame_equal(output, expected_output, check_like=True)

def test_methylation_utils():
    args = MockArgs()

    run_epimetheus(
        pileup = "nanomotif/datasets/geobacillus-plasmids.pileup.bed",
        assembly = "nanomotif/datasets/geobacillus-plasmids.assembly.fasta",
        motifs = ["GATC_m_3", "GATC_a_1"],
        threads = 1,
        min_valid_read_coverage = args.min_valid_read_coverage,
        output = os.path.join(args.out, "motifs-scored-read-methylation.tsv")
    )

    file = Path(os.path.join(args.out, "motifs-scored-read-methylation.tsv"))
    assert file.exists(), "motifs_scored-read-methylation.tsv does not exist"

    res = pl.read_csv(file, separator = "\t")
    print(res.columns)
    assert res.columns == ['contig', 'motif', 'mod_type', 'mod_position', 'median', 'mean_read_cov', 'N_motif_obs', 'motif_occurences_total']
    assert res.shape == (4, 8), "Shape does not match"

    



import logging
import os
from datetime import datetime

def set_logger_config(args):
    """
    Set the configuration for the logger.
    
    params:
        args: args.output - The output directory for the log file
    """
    log_dir = os.path.join(args.out, "log")
    if not os.path.exists(log_dir):
        os.makedirs(log_dir, exist_ok=True)
    
    # Format the current datetime for the log filename
    current_time = datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
    
    logging.basicConfig(
        level=logging.INFO,  # Set to DEBUG, INFO, WARNING, ERROR, or CRITICAL
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        filename=os.path.join(log_dir, f"{current_time}_log.log"),  # Omit or set to None to log to the console
        filemode='w'  # Use 'a' for append; 'w' for overwrite
    )
    
    logging.info("Logger configuration set")
    




import subprocess
import os
import nanomotif
import platform
import polars as pl
import re
import logging

def split_bin_contig(df):
    def custom_replace(bin_compare, contig_bin):
        if bin_compare is not None and contig_bin is not None:
            pattern = re.escape(contig_bin) + "_"
            return re.sub(pattern, "", bin_compare, count=1)
        return bin_compare

    df = df.with_columns(
        pl.when(pl.col("bin_compare").is_not_null()).then(
            pl.struct(["bin_compare", "contig_bin"]).map_elements(
                lambda x: custom_replace(x["bin_compare"], x["contig_bin"])
            , return_dtype = pl.String)
        ).alias("contig")
    )
    return df


def get_bin(bin_contig):
    return re.search(r"(.+)_contig_\d+", bin_contig).group(1)


def add_compare_df(df, bin_contig):
    df = df.filter(pl.col("bin_compare") == bin_contig)\
            .rename({"bin": "contig_bin"})
    
    return df




# pyright: reportUndefinedVariable=false
#%% Load modules
import pandas as pd
from Bio import SeqIO

#%% Importing defenseFinder hmmer and systems output
df_hmmer_path = snakemake.input['hmmer_file']
df_hmmer = pd.read_csv(df_hmmer_path, sep='\t')

df_systems_path = snakemake.input["system_file"]
df_systems = pd.read_csv(df_systems_path, sep='\t')

#%% Filter defensefinder data for RM systems only.
df_hmmer_RM = df_hmmer[df_hmmer['gene_name'].str.contains('_Type_', na=False)]

# Group by gene_id (hit id), then retrieve the entry with the maximum 'hit_score' in each group
idx = df_hmmer_RM.groupby('hit_id')['hit_score'].idxmax()
df_hmmer_RM_filtered = df_hmmer_RM.loc[idx]

#%% Filter defensefinder data for MTase systems only
df_hmmer_MTase = df_hmmer_RM_filtered[(df_hmmer_RM_filtered['gene_name'].str.contains("MTases")) | (df_hmmer_RM_filtered['gene_name'].str.contains("Type_IIG"))].copy()

#%% Determining whether the MTase is part of an RM system
df_systems_RM = df_systems[df_systems['type'].isin(["RM", "MADS", "BREX"])]

# Defining a function to check whether a 'qseqid' is in any 'protein_in_syst' row
def check_RM_system(hit_id, df_systems):
    for index, row in df_systems.iterrows():
        protein_list = row['protein_in_syst'].split(',')
        if hit_id in protein_list:
            return True, row['sys_id']
    return False, None

# Checking whether MTase is part of any system
df_hmmer_MTase[['system', 'sys_id']] = df_hmmer_MTase['hit_id'].apply(lambda x: pd.Series(check_RM_system(x, df_systems)))

# Checking whether MTase is part of an RM-system, MADS-system or BREX-system
df_hmmer_MTase[['RM_system', 'sys_id']] = df_hmmer_MTase["hit_id"].apply(lambda x: pd.Series(check_RM_system(x, df_systems_RM)))

# Removing MTase hits which are part of another defense system 
df_hmmer_MTase_filtered = df_hmmer_MTase[~((df_hmmer_MTase['system'] == True) & (df_hmmer_MTase['RM_system'] == False))].copy()
df_hmmer_MTase_filtered.drop('system', axis = 1, inplace = True)

df_hmmer_MTase_filtered = df_hmmer_MTase_filtered[
    (~df_hmmer_MTase_filtered['gene_name'].str.contains("Type_IIG")) | 
    (df_hmmer_MTase_filtered['gene_name'].str.contains("Type_IIG") & df_hmmer_MTase_filtered['RM_system'] == True)
]

#%% Generating processed defensefinder.tsv
df_hmmer_MTase_filtered.to_csv(snakemake.output['DF_MTase_table'], sep = "\t", index = False)


#%% Generating MTase sequence files
MTase_sequence_file = snakemake.output['DF_MTase_AA']
prodigal_sequence_file = snakemake.input['prodigal_AAs']

# Load identifiers from the DataFrame
identifiers = df_hmmer_MTase_filtered['hit_id'].tolist()

with open(MTase_sequence_file, 'w') as output_file:
    for record in SeqIO.parse(prodigal_sequence_file, 'fasta'):
        if record.id.split(' ')[0] in identifiers:
            SeqIO.write(record, output_file, 'fasta')



import pytest
import polars as pl
from nanomotif.binnary import data_processing, utils

class MockArgs:
    def __init__(self):
        self.pileup = "datasets/binnary_testdata/pileup.bed"
        self.assembly_file = "data/binnary_testdata/assembly_file.fasta"
        self.bin_motifs = "datasets/binnary_testdata/bin-motifs.tsv"
        self.motifs_scored = "datasets/binnary_testdata/motifs-scored.tsv"
        self.contig_bins = "datasets/binnary_testdata/contig_bin.tsv"
        self.min_valid_read_coverage = 3
        self.mean_methylation_cutoff = 0.25
        self.n_motif_bin_cutoff = 500
        self.n_motif_contig_cutoff = 10
        self.ambiguous_motif_percentage_cutoff = 0.40
        self.min_motif_comparisons = 2
        self.save_scores = False
        self.out = "tests/binnary/test_output"
        self.threads = 1

# Remove tests/test_output directory
import shutil
shutil.rmtree("tests/binnary/test_output", ignore_errors=True)


@pytest.fixture(scope="session")
def loaded_data():
    # Instantiate MockArgs
    mock_args = MockArgs()

    # Load the data using the mock_args object
    data = data_processing.load_data(mock_args)

    # Unpack the data tuple to individual variables if needed
    bin_motifs, contig_bins = data

    motifs_scored = pl.read_csv(mock_args.motifs_scored, separator = "\t")

    # Return the data as a dictionary or as individual variables, depending on your preference
    return {
        "motifs_scored": motifs_scored,
        "bin_motifs": bin_motifs,
        "contig_bins": contig_bins
    }
    

@pytest.fixture(scope="session")
def motifs_scored_in_bins_and_bin_motifs(loaded_data):
    args = MockArgs()
    
    # Access the loaded data directly if returned as a dictionary
    motifs_scored = loaded_data["motifs_scored"]
    bin_motifs = loaded_data["bin_motifs"]
    contig_bins = loaded_data["contig_bins"]
    
    bin_motif_binary = data_processing.prepare_bin_consensus(bin_motifs, args)
    
    motifs_in_bins = bin_motif_binary.select("motif_mod").unique()["motif_mod"]
    
    # Step 2: create motifs_scored_in_bins
    motifs_scored_in_bins = data_processing.add_bin(
        motifs_scored, contig_bins 
    )
    
    return {
        "motifs_scored_in_bins": motifs_scored_in_bins,
        "bin_motif_binary": bin_motif_binary
    }



import nanomotif as nm
import polars as pl
import numpy as np
from polars import col
import logging as log

def remove_noisy_motifs(motif_df):
    """
    Remove motifs that have isolated bases
    """
    assert "motif" in motif_df.columns
    assert "mod_position" in motif_df.columns
    assert len(motif_df) > 0
    motif_strings = motif_df.get_column("motif").to_list()
    positions = motif_df.get_column("mod_position").to_list()
    motifs = [nm.candidate.Motif(motif_string, pos) for motif_string, pos in zip(motif_strings, positions)]
    clean_motifs = []
    for motif in motifs:
        if not motif.have_isolated_bases(isolation_size = 3):
            clean_motifs.append(motif.string)
    if not clean_motifs:
        return motif_df.clear()
    else: 
        motif_df_clean = motif_df.filter(pl.col("motif").is_in(clean_motifs))
        return motif_df_clean

def remove_child_motifs(motifs):
    parent_motifs = []
    for i, motif in enumerate(motifs):
        parent = True
        for j, other in enumerate(motifs):
            if i == j:
                continue
            if motif.sub_string_of(other):
                parent = False
                break
        if parent:
            parent_motifs.append(motif)
    return parent_motifs

def remove_sub_motifs(motif_df):
    motif_df_clean = []
    for contig, df in motif_df.groupby("contig"):
        motif_strings = df.get_column("motif").to_list()
        positions = df.get_column("mod_position").to_list()
        motifs = [nm.candidate.Motif(motif_string, pos) for motif_string, pos in zip(motif_strings, positions)]
        parent_motifs = remove_child_motifs(motifs)
        df_clean = df.filter(col("motif").is_in(parent_motifs))
        motif_df_clean.append(df_clean)
    motif_df_clean = pl.concat(motif_df_clean)
    return motif_df_clean

def merge_motifs_in_df(motif_df, pileup, assembly, mean_shift_threshold = -0.2):
    new_df = []
    for (contig, mod_type), df in motif_df.groupby("contig", "mod_type"):
        # Get list of motifs
        motif_seq = df["motif"].to_list()
        motif_pos = df["mod_position"].to_list()
        motifs = [nm.candidate.Motif(seq, pos) for seq, pos in zip(motif_seq, motif_pos)]

        # Merge motifs
        merged_motifs = nm.candidate.merge_motifs(motifs)
        all_merged_motifs = []
        all_premerge_motifs = []
        # Check mean shift of premerge motifs to merged motif is high enough
        for cluster, motifs in merged_motifs.items():
            merged_motif = motifs[0]
            premerge_motifs = motifs[1]
            merge_mean = nm.evaluate.motif_model_contig(
                pileup.filter((col("contig") == contig) & (col("mod_type") == mod_type)), 
                assembly.assembly[contig].sequence,
                merged_motif
            ).mean()
            pre_merge_means = []
            for pre_merge_motif in premerge_motifs:
                pre_merge_means.append(nm.evaluate.motif_model_contig(
                    pileup.filter((col("contig") == contig) & (col("mod_type") == mod_type)), 
                    assembly.assembly[contig].sequence,
                    pre_merge_motif
                ).mean())
            
            pre_merge_mean = sum(np.array(pre_merge_means)) / len(pre_merge_means)
            mean_shift = merge_mean - pre_merge_mean
            if mean_shift < mean_shift_threshold:
                log.info(f"Mean shift of merged motif {merged_motif} is {mean_shift}, keeping original motifs")
            
            else:
                log.info(f"Mean shift of merged motif {merged_motif} is {mean_shift}")
                all_merged_motifs.append(merged_motif)
                all_premerge_motifs.extend(premerge_motifs)

        # Create a new dataframe with the non merged motifs
        if  len(all_premerge_motifs) == 0:
            # No motifs were merged
            new_df.append(df)
            continue
        single_df = df.filter(col("motif").is_in(all_premerge_motifs).not_())
        new_df.append(single_df)
        merged_df = []
        for motif in all_merged_motifs:
            merged_model = nm.evaluate.motif_model_contig(
                pileup.filter((col("contig") == contig) & (col("mod_type") == mod_type)), 
                assembly.assembly[contig].sequence,
                motif
            )
            merged_df.append(pl.DataFrame({
                "sequence": motif.string,
                "score": -1.,
                "contig": contig,
                "mod_type": mod_type,
                "model": merged_model,
                "motif": motif.string,
                "mod_position": motif.mod_position
            }))
        new_df.append(pl.concat(merged_df))
    new_df = pl.concat(new_df)
    return new_df

def join_motif_complements(motif_df):
    if "model" in motif_df.columns:
        if "model" in motif_df.columns:
            n_mod = [x._alpha for x in motif_df["model"]]
            n_nomod = [x._beta for x in motif_df["model"]]

            motif_df = motif_df.with_columns([
                pl.Series("n_mod", n_mod),
                pl.Series("n_nomod", n_nomod)
            ]).drop(["model"])
    motif_df = motif_df.select([
        "contig", "mod_type", "motif", "mod_position", "n_mod", "n_nomod"
    ])
    motif_df = motif_df.with_columns([
        pl.col("motif").apply(lambda x: nm.candidate.regex_to_iupac(x)).alias("motif")
    ]).with_columns([
        pl.col("motif").apply(lambda x: nm.utils.motif_type(x)).alias("motif_type")
    ]).with_columns([
        pl.col("motif").apply(lambda x: nm.candidate.reverse_compliment(x)).alias("motif_complement")
    ])
    
    
    motifs_out = motif_df \
        .join(
            motif_df.select([
                "contig", "mod_type", "motif_complement", "mod_position", "n_mod", "n_nomod"
            ]),
            left_on = ["contig", "mod_type", "motif"],
            right_on = ["contig", "mod_type", "motif_complement"],
            how = "left",
            suffix = "_complement"
        ) \
        .filter(
            (pl.col("motif") >= pl.col("motif_complement")) | pl.col("mod_position_complement").is_null()
        ) \
        .with_columns([
            pl.when(pl.col("mod_position_complement").is_not_null()) \
                .then(pl.col("motif_complement")) \
                .otherwise(None) \
                .alias("motif_complement")
        ])
    return motifs_out


import pytest
import subprocess
import pandas as pd
import os
import glob
import shutil


def test_bin_consensus():
    """
    """
    outdir = "tests/cli/test_bin_consensus"
    
    cmd = [
        "nanomotif", "bin_consensus",
        "-t", "1",
        "nanomotif/datasets/geobacillus-plasmids.assembly.fasta",
        "nanomotif/datasets/geobacillus-plasmids.pileup.bed",
        "nanomotif/datasets/geobacillus-plasmids.motifs.tsv",
        "nanomotif/datasets/geobacillus-contig-bin.tsv",
        "nanomotif/datasets/geobacillus-plasmids.motifs-scored.tsv",
        "--out", outdir
    ]
    
    result = subprocess.run(cmd)
    
    # Check that the CLI tool executed successfully
    assert result.returncode == 0, "CLI tool did not exit successfully"

    #TODO: remove outputdir
    #


def test_score_motifs():
    """
    """
    outdir = "tests/cli/test_score_motifs"
    
    cmd = [
        "nanomotif", "score_motifs",
        "-t", "1",
        "nanomotif/datasets/geobacillus-plasmids.assembly.fasta",
        "nanomotif/datasets/geobacillus-plasmids.pileup.bed",
        "nanomotif/datasets/geobacillus-plasmids.motifs.tsv",
        "--out", outdir
    ]
    
    result = subprocess.run(cmd)
    
    # Check that the CLI tool executed successfully
    assert result.returncode == 0, "CLI tool did not exit successfully"

    #TODO: remove outputdir




import os
import pytest
import pandas as pd
from nanomotif.binnary.data_processing import generate_output
from pathlib import Path

def test_generate_output_new_directory(tmp_path):
    """
    Test generate_output when the output directory does not exist.
    It should create the directory and the output file.
    """
    df = pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]})
    
    output_dir = tmp_path / "new_directory"  # Use Path object for directory
    output_path = output_dir / "output.tsv"  # Full path including filename
    
    generate_output(df, str(output_dir), "output.tsv")
    
    assert output_dir.is_dir(), "Output directory was not created."
    assert output_path.is_file(), "Output file was not created."
    
    result_df = pd.read_csv(output_path, sep="\t")
    pd.testing.assert_frame_equal(result_df, df, check_dtype=False)

def test_generate_output_existing_directory(tmp_path):
    """
    Test generate_output when the output directory already exists.
    """
    df = pd.DataFrame({'col1': [5, 6], 'col2': [7, 8]})
    
    output_path = tmp_path / "output.tsv"  # Path object for the file
    
    generate_output(df, str(tmp_path), "output.tsv")
    
    assert output_path.is_file(), "Output file was not created in the existing directory."
    
    result_df = pd.read_csv(output_path, sep="\t")
    pd.testing.assert_frame_equal(result_df, df, check_dtype=False)


import pytest
from nanomotif.binnary import data_processing as dp
import pandas as pd
import glob
import os

def test_create_contig_bin_contamination(loaded_data):
    """
    Test create_contig_bin function.
    """
    contig_bin = loaded_data["contig_bins"]
    
    contamination = pd.DataFrame({
        'bin': ['b3', 'b3', 'b3'],
        'bin_contig_compare': ['b3_contig_12', 'b3_contig_13', 'b3_contig_6'],
        'binary_methylation_mismatch_score': [2, 2, 3],
        'non_na_comparisons': [4, 4, 4],
        'contig': ['contig_12', 'contig_13', 'contig_6']
    })
        
    new_contig_bin = dp.create_contig_bin_file(contig_bin.to_pandas(), include=None, contamination=contamination)

    assert new_contig_bin is not None, "Output is None"
    # Contigs in contamination may not be in the new contig_bin
    assert not new_contig_bin["contig"].isin(contamination["contig"]).all(), "Contigs in contamination are not in the new contig_bin"



def test_create_contig_bin_include_and_no_contamination(loaded_data):
    """
    Test create_contig_bin function.
    """
    contig_bin = loaded_data["contig_bins"]
    
    # Just for testing purposes it is not used in the function.
    contamination = pd.DataFrame({
        'bin': ['b3', 'b3', 'b3'],
        'bin_contig_compare': ['b3_contig_12', 'b3_contig_13', 'b3_contig_6'],
        'binary_methylation_mismatch_score': [2, 2, 3],
        'non_na_comparisons': [4, 4, 4],
        'contig': ['contig_12', 'contig_13', 'contig_6']
    })
    
    include = pd.DataFrame({
        'bin': ['b2', 'b2', 'b3'],
        'bin_compare': ['b3_contig_6', 'unbinned_contig_14', 'unbinned_contig_15'],
        'binary_methylation_mismatch_score': [0, 0, 0],
        'non_na_comparisons': [4, 4, 2],
        'contig_bin': ['b3', 'unbinned', 'unbinned'],
        'contig': ['contig_6', 'contig_14', 'contig_15']
    })
    contamination_dummy = pd.DataFrame({
        'bin': [],
        'bin_contig_compare': [],
        'binary_methylation_mismatch_score': [],
        'non_na_comparisons': [],
        'contig': []
    })

    new_contig_bin = dp.create_contig_bin_file(contig_bin.to_pandas(), include=include, contamination=contamination_dummy)

    assert new_contig_bin is not None, "Output is None"
    # Contigs in contamination may not be in the new contig_bin
    assert contamination["contig"].isin(new_contig_bin["contig"]).all(), "Contigs in contamination are not in the new contig_bin"
    assert include["contig"].isin(new_contig_bin["contig"]).all(), "Contigs in include are not in the new contig_bin"
    
  

def test_create_contig_bin_include_and_contamination(loaded_data):
    """
    Test create_contig_bin function.
    """
    contig_bin = loaded_data["contig_bins"]
    
    contamination = pd.DataFrame({
        'bin': ['b3', 'b3', 'b3'],
        'bin_contig_compare': ['b3_contig_12', 'b3_contig_13', 'b3_contig_6'],
        'binary_methylation_mismatch_score': [2, 2, 3],
        'non_na_comparisons': [4, 4, 4],
        'contig': ['contig_12', 'contig_13', 'contig_6']
    })
    
    include = pd.DataFrame({
        'bin': ['b2', 'b2', 'b3'],
        'bin_compare': ['b3_contig_6', 'unbinned_contig_14', 'unbinned_contig_15'],
        'binary_methylation_mismatch_score': [0, 0, 0],
        'non_na_comparisons': [4, 4, 2],
        'contig_bin': ['b3', 'unbinned', 'unbinned'],
        'contig': ['contig_6', 'contig_14', 'contig_15']
    })


    new_contig_bin = dp.create_contig_bin_file(contig_bin.to_pandas(), include=include, contamination=contamination)

    assert new_contig_bin is not None, "Output is None"
    # Contigs in contamination may not be in the new contig_bin
    assert not new_contig_bin["contig"].isin(contamination["contig"]).all(), "Contigs in contamination are in the new contig_bin"
    assert include["contig"].isin(new_contig_bin["contig"]).all(), "Contigs in include are not in the new contig_bin"
    
    
def test_write_bins(loaded_data):
    """
    Check if the bins are written correctly
    """
    contig_bin = loaded_data["contig_bins"]
    
    contamination = pd.DataFrame({
        'bin': ['b3', 'b3', 'b3'],
        'bin_contig_compare': ['b3_contig_12', 'b3_contig_13', 'b3_contig_6'],
        'binary_methylation_mismatch_score': [2, 2, 3],
        'non_na_comparisons': [4, 4, 4],
        'contig': ['contig_12', 'contig_13', 'contig_6']
    })
        
    new_contig_bin = dp.create_contig_bin_file(contig_bin.to_pandas(), include=None, contamination=contamination)
    assembly = dp.read_fasta("datasets/binnary_testdata/assembly_file.fasta")
    dp.write_bins_from_contigs(new_contig_bin, assembly, "tests/binnary/bins_test")
    
    bins_in_dir = glob.glob("tests/binnary/bins_test/*.fa")
    assert set(new_contig_bin["bin"]) == set([os.path.basename(x).split(".")[0] for x in bins_in_dir])
    
    contig_ids = []
    with open("tests/binnary/bins_test/b3.fa", 'r') as file:
        for line in file:
            if line.startswith('>'):
                # Extract the contig ID from the description line
                contig_id = line.split()[0][1:]  # Remove '>' and any extra info
                contig_ids.append(contig_id)
    print(contig_ids)
    assert set(contig_ids) == set(new_contig_bin[new_contig_bin["bin"] == "b3"]["contig"])
    
    
    # remove bins 
    for f in bins_in_dir:
        os.remove(f)
    os.rmdir("tests/binnary/bins_test")



import logging
import os

def configure_logger(log_file, verbose=False):
    # Get a logger instance specific to the current process
    logger = logging.getLogger()
    
    # Set the logging level
    logger.setLevel(logging.DEBUG if verbose else logging.INFO)
    
    # Check if the logger already has handlers (to avoid duplicates)
    if not logger.hasHandlers():
        # Create a file handler for logging to a specific file
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(logger.level)
        
        # Set the format for log messages
        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s", datefmt="%d-%b-%y %H:%M:%S")
        file_handler.setFormatter(formatter)
        
        # Add the handler to the logger
        logger.addHandler(file_handler)




import nanomotif.dataload as dl
import polars as pl
import os


def geobacillus_plasmids_pileup_path():
    data_path = os.path.join(os.path.dirname(__file__), "datasets", "geobacillus-plasmids.pileup.bed")
    return data_path

def geobacillus_plasmids_pileup():
    data_path = geobacillus_plasmids_pileup_path()
    pileup = dl.load_pileup(data_path, min_fraction=0.3, min_coverage=5)
    return pileup

def geobacillus_plasmids_assembly_path():
    data_path = os.path.join(os.path.dirname(__file__), "datasets", "geobacillus-plasmids.assembly.fasta")
    return data_path
def geobacillus_plasmids_assembly():
    data_path = geobacillus_plasmids_assembly_path()
    assembly = dl.load_assembly(data_path)
    return assembly

def geobacillus_plasmids_bin_path():
    data_path = os.path.join(os.path.dirname(__file__), "datasets", "geobacillus-contig-bin.tsv")
    return data_path
def geobacillus_plasmids_bin():
    data_path = geobacillus_plasmids_bin_path()
    contig_bin = pl.read_csv(data_path, separator="\t", has_header=False) \
        .rename({"column_1":"contig", "column_2":"bin"})
    return contig_bin






class Pileup():
    """
    Class for loading pileup files
    """
    def __init__(self, pileup):
        self.pileup = pileup

    def get_contigs(self):
        """
        Get contigs in pileup
        """
        return self.pileup["contig"].unique()

    def get_mod_types(self):
        """
        Get modification types in pileup
        """
        return self.pileup["mod_type"].unique()
    
    def get_strands(self):
        """
        Get strands in pileup
        """
        return self.pileup["strand"].unique()
    



# pyright: reportUndefinedVariable=false
#The above line removes warning messages coming from the snakemake.input and .output parameters

#%% Importing packages and utility functions
import pandas as pd
import numpy as np
from utillities import reverse_complement
from utillities import motif_type_predictions
from utillities import RM_type_converter
from utillities import recode_mod_type

#%% Importing modtype table
gene_mod_tsv_path = snakemake.input['mod_table_path']
gene_mod_df = pd.read_csv(gene_mod_tsv_path, sep = '\t', header = 0)


#%% Importing Defensefinder table
DefenseFinder_df_tsv_path = snakemake.input['DefenseFinder_path']
DefenseFinder_df = pd.read_csv(DefenseFinder_df_tsv_path, sep = "\t", header = 0)
DefenseFinder_df = RM_type_converter(DefenseFinder_df, "gene_name") # Converting 'gene name' column to 'sub_type'
DefenseFinder_df = DefenseFinder_df[["hit_id", "replicon", "sub_type", "RM_system", "sys_id"]] 
DefenseFinder_df.columns = ["gene_id", "contig", "sub_type", "RM_system", "sys_id"]


#%% Importing BLASTP table
BLASTP_df_tsv_path = snakemake.input['BLASTP_path']
BLASTP_df = pd.read_csv(BLASTP_df_tsv_path, sep = "\t", header = 0, index_col=False)
BLASTP_df = BLASTP_df[["qseqid", "REbase_id", "motif"]]
BLASTP_df.columns = ['gene_id', 'REbase_ID', 'motif_guess']

DefenseFinder_blastp_df = DefenseFinder_df.merge(BLASTP_df, on = 'gene_id', how = 'left')

#%% Importing nanomotif table
nanomotif_table_path = snakemake.input['nanomotif_table_path']
nanomotif_table = pd.read_csv(nanomotif_table_path, sep = "\t", header = 0)
nanomotif_table['assign_mod_type'] = nanomotif_table['mod_type'].apply(recode_mod_type)

#Set motif acceptance threshold as >=0.5 mean_methylation and remove ambiguous motifs in nanomotif_table
mean_methylation = nanomotif_table['n_mod_bin'] / (nanomotif_table['n_mod_bin'] + nanomotif_table['n_nomod_bin'])
nanomotif_table_mm50 = nanomotif_table[mean_methylation >= 0.5]



#%% Importing bin_contig table from mmlong2 
contig_bin_tsv_path = snakemake.input['contig_bin']
contig_bin_df = pd.read_csv(contig_bin_tsv_path , sep = "\t", header = None)
contig_bin_df.columns = ['contig', 'bin name']


#%% Merging dataframes
# Merging bin origin information into DefenseFinder based on contig number.
bin_DF_blastp_df = DefenseFinder_blastp_df.merge(contig_bin_df[['bin name', 'contig']], on = 'contig', how = 'left')
bin_DF_blastp_df_filtered = bin_DF_blastp_df.dropna(subset=['bin name'])

# Merging modtype table into Defensefinder table
MTase_table = bin_DF_blastp_df_filtered.merge(gene_mod_df[['gene_id', 'mod_type']], on = 'gene_id', how = 'left')


#%% adding methylation motif type column based on MTase subtype
MTase_table = motif_type_predictions(MTase_table, 'sub_type')
MTase_table_assigned = MTase_table.copy()


#%% Add "a" mod type known for RM type I, type IIG and type III
for idx, row in MTase_table_assigned.iterrows():
    if row['sub_type'] == 'Type_I' or row['sub_type'] == 'Type_III' or row['sub_type'] == 'Type_IIG':
        row['mod_type'] = "ac"


#%% Preparing dataframes
        
#Extracting all entries with missing 'mod type' prediction
nan_genes = MTase_table_assigned[MTase_table_assigned['mod_type'].isna()]

# Generate entries with pseudo mod types for all entries with missing 'mod type' prediction
extra_entries = []

for index, entry in nan_genes.iterrows():
        # Create a copy of the current nan gene for 'a' and 'm'
        new_entry_ac = entry.copy()
        new_entry_m = entry.copy()

        # Set 'mod_type' to 'ac' and 'm' respectively
        new_entry_ac['mod_type'] = 'ac'
        new_entry_m['mod_type'] = 'm'


        extra_entries.append(new_entry_ac)
        extra_entries.append(new_entry_m)

extra_entries_df = pd.DataFrame(extra_entries)

# Add entries with pseudo mod types to MTase_table_assigned resulting in combined_MTase
combined_MTase_df = pd.concat([MTase_table_assigned, extra_entries_df]) #Indicies stay the same for extra entries and original entry


# Initialize a column to track whether an entry has been assigned
nanomotif_table_mm50 = nanomotif_table_mm50.copy()
nanomotif_table_mm50.loc[:,'linked'] = False
MTase_table_assigned = MTase_table_assigned.copy()
MTase_table_assigned.loc[:,'linked'] = False


#Group by 'bin name' and 'mod type'
Cgrouped_MTase = combined_MTase_df.groupby(['bin name', 'mod_type'], dropna=True)
grouped_nanomotif = nanomotif_table_mm50.groupby(['bin', 'assign_mod_type'])


#%% Assigment rule system

# Step 1: Priority 1 Assignment based on similarity to motif guess
for (bin_name, assign_mod_type), nanomotif_group in grouped_nanomotif:
    if (bin_name, assign_mod_type) in Cgrouped_MTase.groups:
        mtase_group = Cgrouped_MTase.get_group((bin_name, assign_mod_type))
        for idx, row in nanomotif_group.iterrows():
            # Calculate reverse complement of the motif
            rev_comp_motif = reverse_complement(row['motif'])

            # Find matching motif guess in MTase_table and assign
            matching_indices = mtase_group.loc[mtase_group['motif_guess'] == row['motif']].index
            matching_rev_comp_indices = mtase_group.loc[mtase_group['motif_guess'] == rev_comp_motif].index
            

            # Expand matching_indices to include sys_id matches
            if not matching_indices.empty:
                sys_id_value = mtase_group.loc[matching_indices[0], 'sys_id']  # Get sys_id of the first match
                sys_id_matching_indices = mtase_group.loc[mtase_group['sys_id'] == sys_id_value].index
                matching_indices = matching_indices.union(sys_id_matching_indices)

            # Similarly expand for reverse complement matches
            if not matching_rev_comp_indices.empty:
                sys_id_value_rev_comp = mtase_group.loc[matching_rev_comp_indices[0], 'sys_id']
                sys_id_matching_rev_comp_indices = mtase_group.loc[mtase_group['sys_id'] == sys_id_value_rev_comp].index
                matching_rev_comp_indices = matching_rev_comp_indices.union(sys_id_matching_rev_comp_indices)

            # Update detected motif
            MTase_table_assigned.loc[matching_indices, 'detected_motif'] = row['motif']
            MTase_table_assigned.loc[matching_rev_comp_indices, 'detected_motif'] = rev_comp_motif


            # Mark as linked
            MTase_table_assigned.loc[matching_indices, 'linked'] = True
            MTase_table_assigned.loc[matching_rev_comp_indices, 'linked'] = True

            # If there are matching indices, mark as linked in nanomotif_table
            if not matching_indices.empty:
                nanomotif_group.loc[idx, 'linked'] = True
                genes_str = ', '.join(MTase_table_assigned.loc[matching_indices, 'gene_id'].astype(str).unique())
                nanomotif_table_mm50.at[idx, 'candidate_genes'] = genes_str
            if not matching_rev_comp_indices.empty:
                nanomotif_group.loc[idx, 'linked'] = True
                genes_str = ', '.join(MTase_table_assigned.loc[matching_rev_comp_indices, 'gene_id'].astype(str).unique())
                nanomotif_table_mm50.at[idx, 'candidate_genes'] = genes_str

        # After processing each group, update the 'assigned' status back to the main dataframe
        nanomotif_table_mm50.update(nanomotif_group['linked'])

MTase_table_assigned['detected_motif'] = MTase_table_assigned['detected_motif'].replace('nan', np.nan)
#%%
#%% Re-group to exclude assigned entries
grouped_nanomotif = nanomotif_table_mm50[nanomotif_table_mm50['linked'] != True].groupby(['bin', 'assign_mod_type'])
idx_MTase_table_assigned = MTase_table_assigned[MTase_table_assigned['linked'] != True].index #Extract indicies of non assigned entries 
#%%
combined_MTase_df = combined_MTase_df.loc[idx_MTase_table_assigned] #FIlter combined_MTase based on extracted indicies
Cgrouped_MTase = combined_MTase_df.groupby(['bin name', 'mod_type'], dropna=True)
#%%

# Step 2: Priority 2 Assignment: based single subtypes in group
for (bin_name, assign_mod_type), nanomotif_group in grouped_nanomotif:
    if (bin_name, assign_mod_type) in Cgrouped_MTase.groups:
        mtase_group = Cgrouped_MTase.get_group((bin_name, assign_mod_type))

        # Count unique motif types in nanomotif and MTase table
        for idx, row in nanomotif_group.iterrows():
            motif_type = row['motif_type']
            candidate_genes = mtase_group[mtase_group['motif_type'].apply(lambda x: motif_type in x.split(', '))][['gene_id', 'sys_id']]
            
            # Count unique sys_ids
            unique_sys_ids = candidate_genes['sys_id'].unique()
            if len(unique_sys_ids) == 1 and not pd.isna(unique_sys_ids).any():
                count_in_mtase = 1
            elif len(candidate_genes) == 1 and pd.isna(unique_sys_ids).any():
                count_in_mtase = 1
            else:
                count_in_mtase = len(candidate_genes)

            count_in_nanomotif = nanomotif_group['motif_type'].eq(motif_type).sum()

            # Check if the motif_type is unique in both groups
            if count_in_mtase == 1 and count_in_nanomotif == 1:
                
                #Assigning single candidate genes
                genes_list = candidate_genes['gene_id'].tolist()
                genes_str = ', '.join(genes_list)
                nanomotif_table_mm50.loc[idx, 'candidate_genes'] = genes_str

                if motif_type in ['non-palindrome', 'bipartite']:
                    # Explicitly handle 'bipartite, non-palindrome' if motif_type is 'non-palindrome' or 'bipartite'
                    mtase_index = mtase_group[(mtase_group['motif_type'].eq(motif_type)) | (mtase_group['motif_type'] == 'bipartite, non-palindrome')].index
                else:
                    # Find the corresponding index in mtase_group
                    mtase_index = mtase_group[mtase_group['motif_type'].eq(motif_type)].index
                
                # Assign the motif
                MTase_table_assigned.loc[mtase_index, 'detected_motif'] = row['motif']
                MTase_table_assigned.loc[mtase_index, 'linked'] = True
                nanomotif_table_mm50.loc[idx, 'linked'] = True

MTase_table_assigned['detected_motif'] = MTase_table_assigned['detected_motif'].replace('nan', np.nan)

#%% Re-group to exclude assigned entries
grouped_nanomotif = nanomotif_table_mm50[nanomotif_table_mm50['linked'] != True].groupby(['bin', 'assign_mod_type'])
idx_MTase_table_assigned = MTase_table_assigned[MTase_table_assigned['linked'] != True].index #Extract indicies of non assigned entries
combined_MTase_df = combined_MTase_df.loc[idx_MTase_table_assigned] #Filter combined_MTase based on extracted indicies
Cgrouped_MTase = combined_MTase_df.groupby(['bin name', 'mod_type'], dropna=True)

#%% Step 3: Assign candidate genes
for (bin_name, assign_mod_type), nanomotif_group in grouped_nanomotif:
    if (bin_name, assign_mod_type) in Cgrouped_MTase.groups:
        mtase_group = Cgrouped_MTase.get_group((bin_name, assign_mod_type))

        # Count unique motif types in nanomotif and MTase table
        for idx, row in nanomotif_group.iterrows():
            motif_type = row['motif_type']            
            candidate_genes = mtase_group[mtase_group['motif_type'].apply(lambda x: motif_type in x.split(', '))][['gene_id', 'sys_id']]

            # Count unique sys_ids
            unique_sys_ids = candidate_genes['sys_id'].unique()
            if len(unique_sys_ids) == 1 and not pd.isna(unique_sys_ids).any():
                count_in_mtase = 1
            elif len(candidate_genes) == 1 and pd.isna(unique_sys_ids).any():
                count_in_mtase = 1
            else:
                count_in_mtase = len(candidate_genes)

            count_in_nanomotif = nanomotif_group['motif_type'].eq(motif_type).sum()
            
            # Check if the motif_type is unique in both groups
            if count_in_mtase == 1 and count_in_nanomotif == 1:
                
                #Assigning single candidate genes
                genes_list = candidate_genes['gene_id'].tolist()
                genes_str = ', '.join(genes_list)
                nanomotif_table_mm50.loc[idx, 'candidate_genes'] = genes_str

                if motif_type in ['non-palindrome', 'bipartite']:
                    # Explicitly handle 'bipartite, non-palindrome' if motif_type is 'non-palindrome' or 'bipartite'
                    mtase_index = mtase_group[(mtase_group['motif_type'].eq(motif_type)) | (mtase_group['motif_type'] == 'bipartite, non-palindrome')].index
                else:
                    # Find the corresponding index in mtase_group
                    mtase_index = mtase_group[mtase_group['motif_type'].eq(motif_type)].index
                
                # Assign the motif
                MTase_table_assigned.loc[mtase_index, 'detected_motif'] = row['motif']
                MTase_table_assigned.loc[mtase_index, 'linked'] = True
                nanomotif_table_mm50.loc[idx, 'linked'] = True

            #Assigning all candidate genes
            if count_in_mtase >= 1:
                genes_list = candidate_genes['gene_id'].tolist()
                genes_str = ', '.join(genes_list)

                nanomotif_table_mm50.loc[idx, 'candidate_genes'] = genes_str


# %%
MTase_table_assigned = MTase_table_assigned[['bin name', 'gene_id', 'contig', 'mod_type', 'sub_type', 'RM_system', 'sys_id', 'motif_type', 'REbase_ID', 'motif_guess', 'linked', 'detected_motif']]
MTase_table_assigned.columns = ['bin', 'gene_id', 'contig', 'mod_type_pred', 'sub_type_pred', 'RM_system', 'DF_system_ID', 'motif_type_pred', 'REbase_ID', 'motif_pred', 'linked', 'detected_motif']
MTase_table_assigned_cl = MTase_table_assigned.dropna(subset=['bin'])

nanomotif_table_mm50 = nanomotif_table_mm50[['bin', 'mod_type', 'motif', 'mod_position', 'n_mod_bin', 'n_nomod_bin', 'motif_type', 'motif_complement', 'mod_position_complement', 'n_mod_complement', 'n_nomod_complement', 'linked', 'candidate_genes']]
#%%
MTase_table_assigned_cl.to_csv(snakemake.output['MTase_assignment_table'] , sep='\t', index=False)
nanomotif_table_mm50.to_csv(snakemake.output['nanomotif_assignment_table'] , sep='\t', index=False)




from nanomotif.candidate import *
from nanomotif.constants import *
import nanomotif as nm
import pytest
from hypothesis import given, strategies as st
import itertools
import pandas as pd
import polars as pl
import polars.testing


class TestPileupLoad:
    def test_no_error(self):
        pileup = nm.datasets.geobacillus_plasmids_pileup()
        assert isinstance(pileup.pileup, pl.DataFrame)

    def test_column_types(self):
        pileup = nm.datasets.geobacillus_plasmids_pileup()
        assert pileup.pileup["contig"].dtype == pl.Utf8
        assert pileup.pileup["position"].dtype == pl.Int64
        assert pileup.pileup["mod_type"].dtype == pl.Utf8
        assert pileup.pileup["strand"].dtype == pl.Utf8
        assert pileup.pileup["fraction_mod"].dtype == pl.Float64
        assert pileup.pileup["Nvalid_cov"].dtype == pl.Int64

    def test_fraction_filter(self):
        pileup_path = nm.datasets.geobacillus_plasmids_pileup_path()
        pileup = nm.dataload.load_pileup(pileup_path, min_fraction=0.5, min_coverage=5)
        assert (pileup.pileup["fraction_mod"] > 0.5).all()
        # Assert that the filtered pileup is a subset of the complete pileup
        pileup_complete = nm.dataload.load_pileup(pileup_path, min_fraction=0.5, min_coverage=5)
        assert len(pileup.pileup.join(pileup_complete.pileup, on=pileup.pileup.columns, how="inner")) == len(pileup.pileup)







def identify_motifs(methylation_sequences, contig, pileup, 
                    min_sequences = 50, min_cdf_score = 0.5, cdf_limit = 0.55, 
                    n_contig_samples = 10000, canonical_base = "A", min_kl_divergence = 0.1):
    """
    Identify candidate methylation motifs in a contig

    Parameters:
    - methylation_sequences (EqualLengthDNASet): The methylation sequences to be processed.
    - contig (str): The contig to be processed.
    - pileup (Pileup): The pileup to be processed.
    - min_sequences (int): The minimum number of methylation sequences.
    - min_cdf_score (float): The minimum score of 1 - cdf(cdf_limit) for a motif to be considered valid.
    - cdf_limit (float): The position to evaluate the cdf at.
    - n_contig_samples (int): The number of samples to take from the contig.
    - canonical_base (str): The canonical base of methylation (6mA = A).
    - min_kl_divergence (float): Early stopping criteria, if max KL-divergence falls below, stops building motif.

    Returns:
    - tuple: A tuple of three lists, the first containing the candidate motifs, the second containing the models for the candidates, the third containing the scores for the candidates.
    """
    # Infer padding from sequence length
    padding = methylation_sequences.sequence_length // 2

    # Sample sequence in contig to get background for KL-divergence
    contig_sequences = contig.sample_n_at_subsequence(padding, canonical_base, n_contig_samples)
    contig_pssm = contig_sequences.pssm()

    # Initialize candidate lists
    candidate_motifs = []
    failed_canidates = []
    candidate_models = []
    candidate_scores = []

    # Bases in the DNA sequences (To make sure the order is the same as used internally in the class)
    bases = methylation_sequences.sequences[0].bases

    # Convert DNA sequences to numpy array
    methylation_sequences = convert_seqeunces_to_numpy(methylation_sequences.sequences)

    while True:
        active_methylation_sequences = methylation_sequences.copy()

        if len(candidate_motifs) > 0:
            # Remove all previously identified candidates from sequence set
            for motif in candidate_motifs:
                active_methylation_sequences = subset_DNA_array(active_methylation_sequences, motif, remove_matches = True)
        if len(failed_canidates) > 0:
            # Remove all previously failed candidates from sequence set
            for motif in failed_canidates:
                active_methylation_sequences = subset_DNA_array(active_methylation_sequences, motif, remove_matches = True)

        if active_methylation_sequences.shape[0] < min_sequences:
            break
        
        # Initialize the active candidate as blank sequence
        active_candidate = ["."] * padding + [canonical_base] + ["."] * padding
        evaluated_positions = []
        candidates = []
        best_candidate = None

        while True:
            # Calculate KL-divergence
            methylation_pssm = calculate_pssm(active_methylation_sequences)
            kl_divergence = entropy(methylation_pssm, contig_pssm)

            # Set KL-divergence to 0 for already evaluated positions
            kl_divergence[evaluated_positions] = 0

            # Get the index of the maximum KL-divergence
            max_kl_index = kl_divergence.argmax()

            # Update the active candidate
            new_base = [bases[i] for i in np.argwhere(methylation_pssm[:, max_kl_index] > 0.45)[:, 0]]
            new_base = "".join(new_base)

            # If multiple bases have a high probability, use a regex expression to represent: [bases]
            if len(new_base) == 0:
                new_base = bases[methylation_pssm[:, max_kl_index].argmax()]
            if len(new_base) > 1:
                new_base = "[" + new_base + "]"
            active_candidate[max_kl_index] = new_base

            model = score_candidate(pileup, contig.sequence, "".join(active_candidate), padding)
            score = 1 - model.cdf(cdf_limit)

            evaluated_positions.append(max_kl_index)
            log.debug(f"{''.join(active_candidate)} | cdf score: {score:.3f} | mean: {model.mean():.3f} | {model} | max kl: {kl_divergence[max_kl_index]:.3f}")
            
            active_methylation_sequences = subset_DNA_array(active_methylation_sequences, active_candidate, remove_matches = False)

            candidates.append((active_candidate, model, score))

            # Success criteria
            if score >= min_cdf_score:
                failed = False
                best_candidate = len(candidates) - 1
                break

            # Failure criteria
            if active_methylation_sequences.shape[0] < min_sequences:
                failed = True
                log.debug("Too few sequences left")
                break
            if len(evaluated_positions) >= padding * 2 + 1:
                failed = True
                log.debug("Too many positions evaluated")
                break
            if kl_divergence[max_kl_index] < min_kl_divergence:
                failed = True
                log.debug("Low KL divergence")
                break

        if failed:
            failed_canidates.append(active_candidate)
        else:
            log.debug("Saving candidate")
            candidate_motifs.append(candidates[best_candidate][0])
            candidate_models.append(candidates[best_candidate][1])
            candidate_scores.append(candidates[best_candidate][2])
    candidate_motifs_str = ["".join(motif) for motif in candidate_motifs]
    return candidate_motifs_str, candidate_models, candidate_scores




import polars as pl
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.neighbors import KNeighborsClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.decomposition import PCA
from nanomotif.binnary import data_processing as dp 
from nanomotif.binnary import utils as ut
import logging



def include_contigs(contig_methylation, contig_lengths, mean_probability):
    logger = logging.getLogger(__name__)
    logger.info("Starting include_contigs analysis...")
    
    contig_bin = contig_methylation\
        .select(["bin", "contig"])\
        .unique()
    
    binned_contig_methylation = dp.impute_contig_methylation_within_bin(
        contig_methylation
    )


    binned_contig_names, binned_matrix = dp.create_matrix(binned_contig_methylation)

    pca_variance=0.90
    pca = PCA(n_components=pca_variance, svd_solver = "full")
    binned_matrix = pca.fit_transform(binned_matrix)
    
    
    bins = pl.DataFrame({
                            "contig": binned_contig_names,
                        })\
            .join(contig_bin, on= "contig", how = "left")


    unbinned_contig_methylation = dp.impute_unbinned_contigs(contig_methylation)
    unbinned_contig_methylation_feature_completions = pl.DataFrame({
                                                           "contig": unbinned_contig_methylation.get_column("contig").unique()
                                                       })\
            .join(
                pl.DataFrame({"motif_mod": binned_contig_methylation.get_column("motif_mod").unique()}), how = "cross"
            )\
            .join(unbinned_contig_methylation, on = ["contig", "motif_mod"], how = "left")
            

    
    unbinned_contig_names, unbinned_matrix = dp.create_matrix(unbinned_contig_methylation_feature_completions)
    unbinned_matrix = pca.transform(unbinned_matrix)

    knn = KNeighborsClassifier(n_neighbors = 3)
    rf = RandomForestClassifier(n_estimators=100, random_state=42)
    lda = LinearDiscriminantAnalysis()

    lda_pred = lda.fit(binned_matrix, bins.get_column("bin"))
    lda_labels = lda_pred.predict(unbinned_matrix)
    lda_prob = lda_pred.predict_proba(unbinned_matrix).max(axis = 1)
    
    knn_pred = knn.fit(binned_matrix, bins.get_column("bin"))
    knn_labels = knn_pred.predict(unbinned_matrix)
    knn_prob = knn_pred.predict_proba(unbinned_matrix).max(axis = 1)
    
    rf_pred = rf.fit(binned_matrix, bins.get_column("bin"))
    rf_labels = rf_pred.predict(unbinned_matrix)
    rf_prob = rf_pred.predict_proba(unbinned_matrix).max(axis = 1)
    
    rf_df = pl.DataFrame({
        "contig": unbinned_contig_names,
        "method": "rf",
        "pred": rf_labels,
        "prob": rf_prob
    })

    knn_df = pl.DataFrame({
        "contig": unbinned_contig_names,
        "method": "knn",
        "pred": knn_labels,
        "prob": knn_prob
    })

    lda_df = pl.DataFrame({
        "contig": unbinned_contig_names,
        "method": "lda",
        "pred": lda_labels,
        "prob": lda_prob
    })

    prob_df = pl.concat([rf_df, knn_df, lda_df])
    
    results = pl.DataFrame({
                    "contig": unbinned_contig_names,
                    "lda": lda_labels,
                    "knn": knn_labels,
                    "rf": rf_labels,
                })\
                .join(contig_bin, on="contig")\
                .melt(
                    id_vars = ["contig", "bin"],
                    value_vars=["lda", "knn", "rf"],
                    value_name="assigned_bin",
                    variable_name="method"
                )
    
    columns = ["contig", "bin", "assigned_bin", "method", "prob", "mean_prob", "confidence"]
    n_assigned_bins = results\
        .group_by(["contig"])\
        .agg(
            pl.col("assigned_bin").n_unique().alias("n_assigned_bins")
        )

    high_confidence_assignements = results\
    .filter(pl.col("contig").is_in(n_assigned_bins.filter(pl.col("n_assigned_bins") == 1).get_column("contig")))


    high_mean_prob = high_confidence_assignements\
        .join(prob_df, on = ["contig", "method"])\
        .group_by(["contig"])\
        .agg(
            pl.col("prob").mean().alias("mean_prob")
        )


    high_confidence_assignements = high_confidence_assignements\
        .join(prob_df, on = ["contig", "method"])\
        .join(high_mean_prob, on = ["contig"])\
        .with_columns(
            pl.when(pl.col("mean_prob") >= mean_probability).then(pl.lit("high_confidence")).otherwise(pl.lit("medium_confidence")).alias("confidence")
        )\
        .sort(["contig"])\
        .select(columns)

    low_confidence_assignments = results\
    .filter(pl.col("contig").is_in(n_assigned_bins.filter(pl.col("n_assigned_bins") == 2).get_column("contig")))

    
    low_confidence_classification = low_confidence_assignments\
        .group_by(["contig", "assigned_bin"])\
        .agg(
            pl.col("assigned_bin").count().alias("n_assigned_bins"),
        )\
        .filter(pl.col("n_assigned_bins") == 2)\
        .drop(["n_assigned_bins"])


    low_confidence_assignments = low_confidence_classification\
        .join(low_confidence_assignments, on = ["contig", "assigned_bin"])\
        .join(prob_df, on = ["contig", "method"])
        

    low_mean_prob = low_confidence_assignments\
        .group_by(["contig"])\
        .agg(
            pl.col("prob").mean().alias("mean_prob")
        )

    low_confidence_assignments = low_confidence_assignments\
        .join(low_mean_prob, on = ["contig"])\
        .with_columns(
            pl.lit("low_confidence").alias("confidence")
        )\
        .sort(["contig"])\
        .select(columns)
    assigned_contigs = pl.concat([high_confidence_assignements, low_confidence_assignments])\
        .sort(["confidence", "contig"])
    return assigned_contigs



import nanomotif as nm
from nanomotif.logger import configure_logger
import logging as log
import os
import sys
import shutil
from pathlib import Path
import json

import numpy as np
import random
import warnings
from nanomotif._version import __version__


def shared_setup(args, working_dir):
    # Set up logging
    LOG_DIR = working_dir + "/logs"
    Path(LOG_DIR).mkdir(parents=True, exist_ok=True)
    configure_logger(LOG_DIR + f"/{args.command}.main.log", args.verbose)
    warnings.filterwarnings("ignore")
    # Check if output directory exists
    if not os.path.exists(args.out):
        os.makedirs(args.out)
    else:
        log.warning(f"Output directory {args.out} already exists")


    # Log arguments
    log.info(f"nanomotif version: {__version__}")
    log.debug("Arguments:")
    for arg in vars(args):
        log.debug(f"{arg}: {getattr(args, arg)}")
    
    # Save settings
    with open(working_dir + f"/args.{args.command}.json", "w") as f:
        json.dump(vars(args), f, indent=2)
    
    # Set seeds
    nm.seed.set_seed(args.seed)




def find_motifs(args, pl,  pileup = None, assembly = None, min_mods_pr_contig = 50, min_mod_frequency = 10000):
    """
    Nanomotif motif finder module

    Args:
        args (argparse.Namespace): Arguments
        pileup (pandas.DataFrame): Pileup data
        assembly (nanomotif.Assembly): Assembly data
    
    Returns:
        pandas.DataFrame: Motif data
    """
    import polars as pl

    log.info("Starting nanomotif motif finder")
    if pileup is None:
        log.info("Loading pileup")
        if not os.path.exists(args.pileup):
            log.error(f"File {args.pileup} does not exist")
            return None
        if args.read_level_methylation:
            pileup = nm.load_pileup(args.pileup, min_coverage = args.threshold_valid_coverage, min_fraction = 0)
        else:
            pileup = nm.load_pileup(args.pileup, min_coverage = args.threshold_valid_coverage, min_fraction = args.threshold_methylation_general)
    

    # Assembly
    if assembly is None:
        log.info("Loading assembly")
        assembly = nm.load_assembly(args.assembly)
    
    pileup = pileup.pileup.with_columns([
        (pl.col("contig") + "_" + pl.col("mod_type")).alias("contig_mod")
    ])
    contig_mods_to_keep, contig_mods_to_remove = nm.dataload.extract_contig_mods_with_sufficient_information(pileup, assembly, min_mods_pr_contig, min_mod_frequency)
    if len(contig_mods_to_keep) == 0:
        log.info("No contigs with sufficient information")
        return None

    log.debug(f"Filtering pileup to keep contigs with more than {min_mods_pr_contig} mods and mod frequency of 1 pr. {min_mod_frequency}")
    pileup = pileup.filter(pl.col("contig_mod").is_in(contig_mods_to_keep))

    # Load low coverage positions
    log.info("Loading low coverage positions")
    low_coverage_positions = nm.load_low_coverage_positions(args.pileup, contig_mods_to_keep, min_coverage = args.threshold_valid_coverage)
    
    # Writing temperary files
    os.makedirs(args.out + "/temp/", exist_ok=True)
    pileup.write_csv(args.out + "/temp/filtered_pileup.tsv", separator="\t")
    with open(args.out + '/temp/contig_mod_combinations_not_processed.tsv', 'w') as f:
        for line in contig_mods_to_remove:
            f.write(f"{line}\n")
    with open(args.out + '/temp/contig_mod_combinations_processed.tsv', 'w') as f:
        for line in contig_mods_to_keep:
            f.write(f"{line}\n")

    log.info("Identifying motifs")
    motifs = nm.evaluate.process_sample_parallel(
            assembly, pileup, 
            low_coverage_positions = low_coverage_positions,
            read_level_methylation = args.read_level_methylation,
            threads = args.threads,
            search_frame_size = args.search_frame_size,
            threshold_methylation_confident = args.threshold_methylation_confident,
            threshold_methylation_general = args.threshold_methylation_general,
            minimum_kl_divergence = args.minimum_kl_divergence,
            verbose = args.verbose,
            log_dir = args.out + "/logs",
            seed = args.seed
        )
    motifs = pl.DataFrame(motifs)
    if motifs is None or len(motifs) == 0:
        log.info("No contigs with sufficient modifications")
        return

    log.info("Writing motifs")
    def format_motif_df(df):
        if "model" in df.columns:
            n_mod = [x._alpha for x in df["model"]]
            n_nomod = [x._beta for x in df["model"]]
        motif_iupac = [nm.seq.regex_to_iupac(x) for x in df["motif"]]
        motif_type = [nm.utils.motif_type(x) for x in motif_iupac]

        df_out = df.with_columns([
            pl.Series("motif", motif_iupac),
            pl.Series("motif_type", motif_type)
        ])
        if "model" in df.columns:
            df_out = df_out.with_columns([
                pl.Series("n_mod", n_mod),
                pl.Series("n_nomod", n_nomod)
            ])
        try:
            df_out = df_out.select([
                "contig", "motif", "mod_position", "mod_type", "n_mod", "n_nomod", "motif_type",
                "motif_complement", "mod_position_complement", "n_mod_complement", "n_nomod_complement"
            ])
        except:
            df_out = df_out.select([
                "contig", "motif", "mod_position", "mod_type", "n_mod", "n_nomod", "motif_type",
            ])
        # Subtract prior alpha and beta from n_mod and n_nomod
        if "model" in df.columns:
            df_out = df_out.with_columns([
                pl.col("n_mod") - nm.model.DEFAULT_PRIOR_BETA,
                pl.col("n_nomod") - nm.model.DEFAULT_PRIOR_ALPHA
            ])
        return df_out
    def save_motif_df(df, name):
        df = format_motif_df(df)
        df = df.sort(["contig", "mod_type", "motif"])
        df.write_csv(args.out + "/" + name + ".tsv", separator="\t")
    os.makedirs(args.out + "/precleanup-motifs/", exist_ok=True)
    motifs.drop("model").write_csv(args.out + "/precleanup-motifs/motifs-raw-unformatted.tsv", separator="\t")
    save_motif_df(motifs, "precleanup-motifs/motifs-raw")

    log.info("Postprocessing motifs")
    motifs_file_name = "precleanup-motifs/motifs"

    log.info(" - Writing motifs")
    motifs = motifs.filter(pl.col("score") > args.min_motif_score)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    
    motifs_file_name = motifs_file_name + "-score"
    save_motif_df(motifs, motifs_file_name)

    log.info(" - Removing sub motifs")
    motifs = nm.postprocess.remove_sub_motifs(motifs)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    motifs_file_name = motifs_file_name +   "-sub"
    save_motif_df(motifs, motifs_file_name)

    log.info(" - Removing noisy motifs")
    motifs = nm.postprocess.remove_noisy_motifs(motifs)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    motifs_file_name = motifs_file_name +   "-noise"
    save_motif_df(motifs, motifs_file_name)

    log.info(" - Merging motifs")
    motifs = nm.postprocess.merge_motifs_in_df(motifs, pileup, assembly)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    motifs_file_name = motifs_file_name +   "-merge"
    save_motif_df(motifs, motifs_file_name)

    log.info(" - Joining motif complements")
    motifs = nm.postprocess.join_motif_complements(motifs)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    motifs_file_name = motifs_file_name +   "-complement"
    save_motif_df(motifs, motifs_file_name)

    log.info(" - Removing motifs observed less than min count")
    motifs = motifs.filter((pl.col("n_mod") + pl.col("n_nomod")) > args.min_motifs_contig)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    motifs_file_name = motifs_file_name +   "-mincount"
    save_motif_df(motifs, motifs_file_name)



    save_motif_df(motifs, "motifs")

    log.info("Done finding motifs")
    return format_motif_df(motifs)

def score_motifs(args, pl, pileup = None, assembly = None, motifs = None, min_mods_pr_contig = 50, min_mod_frequency = 10000):
    log.info("Starting nanomotif motif scorer")
    if pileup is None:
        log.info("Loading pileup")
        pileup =  nm.load_pileup(args.pileup, min_fraction = args.threshold_methylation_general, min_coverage = args.threshold_valid_coverage)
    if assembly is None:
        log.info("Loading assembly")
        assembly = nm.load_assembly(args.assembly)
    if motifs is None:
        log.info("Loading motifs")
        motifs = pl.read_csv(args.motifs, separator="\t")
    if args.save_motif_positions:
        os.makedirs(args.out + "/motif-positions", exist_ok=True)

    pileup = pileup.pileup.filter(pl.col("fraction_mod") > args.threshold_methylation_general)
    # Ensure motif are iupac
    motifs.with_columns([
        pl.col("motif").map_elements(lambda x: nm.seq.regex_to_iupac(x)).alias("motif")
    ])

    
    if any([item for item in motifs.columns if "_complement" in item]):
        log.debug("Reformatting motifs to include scoring of complementary motifs")
        motifs_fwd = motifs[[item for item in motifs.columns if "complement" not in item]]
        motifs_rev = motifs[["contig", "motif_type", "mod_type"] + [item for item in motifs.columns if "complement" in item]] \
            .select(pl.all().name.map(lambda col_name: col_name.replace('_complement', ''))) \
            .select(motifs_fwd.columns) \
            .filter(pl.col("motif").is_not_null())
        
        motifs = pl.concat([motifs_fwd, motifs_rev]).unique(["motif", "contig", "mod_type", "mod_position"])
    
    # Convert to regex for motif scoring
    motifs.with_columns([
        pl.col("motif").map_elements(lambda x: nm.seq.iupac_to_regex(x)).alias("motif")
    ])

    pileup = pileup.with_columns([
        (pl.col("contig") + "_" + pl.col("mod_type")).alias("contig_mod")
    ])
    contig_mods_to_keep, contig_mods_to_remove = nm.dataload.extract_contig_mods_with_sufficient_information(pileup, assembly, min_mods_pr_contig, min_mod_frequency)
    log.debug(f"Filtering pileup to keep contigs with more than {min_mods_pr_contig} mods and mod frequency of 1 pr. {min_mod_frequency}")
    pileup = pileup.filter(pl.col("contig_mod").is_in(contig_mods_to_keep))

    # Load low coverage positions
    log.info("Loading low coverage positions")
    na_position = nm.load_low_coverage_positions(args.pileup, contig_mods_to_keep, min_coverage = args.threshold_valid_coverage)
    

    log.info("Scoring motifs")
    scored_all = nm.scoremotifs.score_sample_parallel(
        assembly, pileup, motifs,
        na_position = na_position,
        threads = args.threads,
        threshold_methylation_general = args.threshold_methylation_general,
        threshold_valid_coverage = 1,
        verbose = args.verbose,
        log_dir = args.out + "/logs",
        save_motif_positions = args.save_motif_positions,
        positions_outdir = args.out + "/motif-positions",
        seed = args.seed
        )
    scored_all = nm.postprocess.join_motif_complements(scored_all)

    scored_all = scored_all.unique([
            "motif", "contig", "mod_type", "mod_position"
        ])
    
    scored_all = scored_all.sort(["contig", "mod_type", "motif"])
    scored_all.write_csv(args.out + "/motifs-scored.tsv", separator="\t")
    return scored_all

def bin_consensus(args, pl, pileup = None, assembly = None, motifs = None, motifs_scored = None):
    bins = pl.read_csv(args.bins, separator="\t", has_header=False, infer_schema_length=10000) \
        .rename({"column_1":"contig", "column_2":"bin"})
    if motifs is None:
        log.info("Loading motifs-scored")
        motifs = pl.read_csv(args.motifs, separator="\t")
    if motifs_scored is None:
        log.info("Loading motifs")
        motifs_scored = pl.read_csv(args.motifs_scored, separator="\t")
    if pileup is None:
        log.info("Loading pileup")
        pileup =  nm.load_pileup(args.pileup, min_fraction = args.threshold_methylation_general, min_coverage = args.threshold_valid_coverage)
    if assembly is None:
        log.info("Loading assembly")
        assembly = nm.load_assembly(args.assembly)

    
    if any([item for item in motifs_scored.columns if "_complement" in item]):
        motifs_fwd = motifs_scored[[item for item in motifs_scored.columns if "complement" not in item]]
        motifs_rev = motifs_scored[["contig", "motif_type", "mod_type"] + [item for item in motifs_scored.columns if "complement" in item]] \
            .select(pl.all().name.map(lambda col_name: col_name.replace('_complement', ''))) \
            .select(motifs_fwd.columns) \
            .filter(pl.col("motif").is_not_null())
        motifs_scored = pl.concat([motifs_fwd, motifs_rev]).unique(["motif", "contig", "mod_type", "mod_position"])
    
    if any([item for item in motifs.columns if "_complement" in item]):
        motifs_fwd = motifs[[item for item in motifs.columns if "complement" not in item]]
        motifs_rev = motifs[["contig", "motif_type", "mod_type"] + [item for item in motifs.columns if "complement" in item]] \
            .select(pl.all().name.map(lambda col_name: col_name.replace('_complement', ''))) \
            .select(motifs_fwd.columns) \
            .filter(pl.col("motif").is_not_null())
        motifs = pl.concat([motifs_fwd, motifs_rev]).unique(["motif", "contig", "mod_type", "mod_position"])


    output = nm.bin_consensus.within_bin_motifs_consensus(pileup, assembly, motifs, motifs_scored, bins)
    output = nm.bin_consensus.merge_bin_motifs(output, bins, pileup, assembly)

    output = output.rename({"bin":"contig", "n_mod_bin":"n_mod", "n_nomod_bin":"n_nomod"})

    output = nm.postprocess.join_motif_complements(output)

    output = output.rename({"contig":"bin", "n_mod":"n_mod_bin", "n_nomod":"n_nomod_bin"})
    output = output.sort(["bin", "mod_type", "motif"])

    output = output.filter(pl.col("n_mod_bin") + pl.col("n_nomod_bin") > args.min_motifs_bin)
    # Ensure no complete duplicates are present
    output = output.unique(["motif", "bin", "mod_type", "mod_position", "n_mod_bin", "n_nomod_bin"])
    output.write_csv(args.out + "/bin-motifs.tsv", separator="\t")

def motif_discovery(args, pl):
    # Check if all required files exist
    if not os.path.exists(args.pileup):
        log.error(f"File {args.pileup} does not exist")
        return
    if not os.path.exists(args.assembly):
        log.error(f"File {args.assembly} does not exist")
        return
    if not os.path.exists(args.bins):
        log.error(f"File {args.bins} does not exist")
        return
    empty_find_motifs = pl.DataFrame({
        "contig": [],
        "motif": [],
        "mod_position": [],
        "mod_type": [],
        "n_mod": [],
        "n_nomod": [],
        "motif_type": [],
        "motif_complement": [],
        "mod_position_complement": [],
        "n_mod_complement": [],
        "n_nomod_complement": []
    })
    empty_bin_motifs = pl.DataFrame({
        "bin": [],
        "mod_type": [],
        "motif": [],
        "mod_position": [],
        "n_mod_bin": [],
        "n_nomod_bin": [],
        "motif_type": [],
        "motif_complement": [],
        "mod_position_complement": [],
        "n_mod_complement": [],
        "n_nomod_complement": []
    })
    
    # Check if output directory exists
    log.info("Loading required files")
    log.debug("Loading pileup")
    if args.read_level_methylation:
        pileup = nm.load_pileup(args.pileup,min_coverage = args.threshold_valid_coverage, min_fraction = 0)
    else:
        pileup = nm.load_pileup(args.pileup,min_coverage = args.threshold_valid_coverage, min_fraction = args.threshold_methylation_general)
        log.debug("Loading assembly")
    assembly = nm.load_assembly(args.assembly)

    # Find motifs
    log.info("Finding motifs")
    motifs = find_motifs(args, pl, pileup=pileup, assembly=assembly)
    if motifs is None:
        empty_find_motifs.write_csv(args.out + "/motifs.tsv", separator="\t")
        empty_bin_motifs.write_csv(args.out + "/bin-motifs.tsv", separator="\t")
        log.info("Stopping workflow")
        return

    # Score all motifs
    log.info("Scoring motifs")
    scored_all = score_motifs(args, pl, pileup=pileup, assembly=assembly, motifs=motifs)

    # Bin consensus
    log.info("Finding bin consensus motifs")
    bin_consensus(args, pl, pileup=pileup, assembly=assembly, motifs=motifs, motifs_scored=scored_all)

    log.info("Done")

def check_install(args, pl):
    
    # Check if output directory exists
    log.info("Loading required files")
    args.out = "nanomotif_install_check"
    args.save_motif_positions = False
    args.pileup = nm.datasets.geobacillus_plasmids_pileup_path()

    pileup = nm.datasets.geobacillus_plasmids_pileup()
    assembly = nm.datasets.geobacillus_plasmids_assembly()

    # Find motifs
    log.info("Finding motifs")
    motifs = find_motifs(args, pl, pileup=pileup, assembly=assembly)

    # Score all motifs
    log.info("Scoring motifs")
    scored_all = score_motifs(args, pl, pileup=pileup, assembly=assembly, motifs=motifs)

    # Bin consensus
    log.info("Finding bin consensus motifs")
    args.bins = nm.datasets.geobacillus_plasmids_bin_path()
    bin_consensus(args, pl, pileup=pileup, assembly=assembly, motifs=motifs, motifs_scored=scored_all)
    
    log.info("Done")
    for i in range(3):
        try:
            shutil.rmtree(args.out)
            break
        except OSError as e:
            time.sleep(0.5) 


####################################################################################################
# Binnary - contamination and inclusion
from nanomotif.binnary import data_processing, detect_contamination, include_contigs
from nanomotif.binnary.logging import set_logger_config
from pymethylation_utils.utils import run_epimetheus


def binnary(args, pl):
    """
    binnary entry point for the DNA Methylation Pattern Analysis tool.
    Orchestrates the workflow of the tool based on the provided arguments.
    """
    
    print("Starting Binnary ", args.command, " analysis...")

    
    # Settting up the logger
    log.info("Starting Binnary analysis...")
    
    # Step 1: Load and preprocess data
    # These functions would be defined in your data_processing module
    (
        bin_motifs,
        contig_bins,
    ) = data_processing.load_data(args)

    

    # Step 2: create motifs_scored_in_bins and bin_consensus_motifs_filtered
    motifs_in_bin_consensus = bin_motifs\
        .with_columns(
            (pl.col("motif") + "_" + pl.col("mod_type") + "_" + pl.col("mod_position").cast(pl.Utf8)).alias("motif_mod")            
        )\
        .get_column("motif_mod").unique()

    contig_methylation_file = "motifs-scored-read-methylation.tsv"
    if not args.force:
        log.info(f"Check if {contig_methylation_file} exists")

    if os.path.isfile(os.path.join(args.out,contig_methylation_file)) and not args.force:
        log.info("motifs-scored-read-methylation.tsv exists. Using existing file! Use --force to override this.")
    elif not os.path.isfile(os.path.join(args.out, contig_methylation_file)) or args.force:
        log.info(f"Running epimetheus to create {contig_methylation_file}")
        # Create motifs-scored-read-methylation
        return_code = run_epimetheus(
            pileup = args.pileup,
            assembly = args.assembly,
            motifs = motifs_in_bin_consensus,
            threads = args.threads,
            min_valid_read_coverage = args.min_valid_read_coverage,
            output = os.path.join(args.out,contig_methylation_file)
        )

        if return_code != 0:
            log.error("Error running epimetheus")


    log.info("Loading assembly file...")
    assembly = data_processing.read_fasta(args.assembly)
    contig_lengths = data_processing.find_contig_lengths(assembly) 

    # Load motifs-scored-read-methylation.tsv
    contig_methylation = pl.read_csv(
        os.path.join(args.out, contig_methylation_file), separator="\t", has_header = True
    )

    contig_methylation = contig_methylation\
        .filter((pl.col("N_motif_obs").cast(pl.Float64) * pl.col("mean_read_cov")) >= args.methylation_threshold)

    # Setting up the contamination analysis
    if (args.command == "detect_contamination" and not args.contamination_file) or (args.command == "include_contigs" and args.run_detect_contamination):
    
        contig_methylation_cont = data_processing.add_bin(
            contig_methylation,
            contig_bins
        )

        contamination = detect_contamination.detect_contamination(
            contig_methylation_cont, contig_lengths, args.num_consensus, args.threads
        )


        data_processing.generate_output(contamination.to_pandas(), args.out, "bin_contamination.tsv")

    if args.command == "detect_contamination":
        # Create a decontaminated contig_bin file
        new_contig_bins = data_processing.create_contig_bin_file(
            contig_bins=contig_bins.to_pandas(), 
            contamination=contamination.to_pandas(),
            include=None
        )
        data_processing.generate_output(new_contig_bins, args.out, "decontaminated_contig_bin.tsv")

    if args.command == "include_contigs":
        # User provided contamination file
        if args.contamination_file:
            log.info("Loading contamination file...")
            contamination = data_processing.load_contamination_file(args.contamination_file)
        # If run_detect_contamination is false and contamination file is not provided, then set contamination to None
        if not args.run_detect_contamination and not args.contamination_file:
            contamination = pl.DataFrame(
                {
                    "contig": []
                }
            )
        
        # move contigs in bin_contamination to unbinned
        contamination_contigs = contamination.get_column("contig")

        log.info("Removing contaminants from bins")
        contig_bins = contig_bins.filter(~pl.col("contig").is_in(contamination_contigs))
        
        contig_methylation_inc = data_processing.add_bin(
            contig_methylation,
            contig_bins
        )

        # Run the include_contigs analysis    
        include_contigs_df = include_contigs.include_contigs(
            contig_methylation_inc,
            contig_lengths,
            mean_probability = args.mean_model_confidence
        )
        
        # Save the include_contigs_df results
        data_processing.generate_output(include_contigs_df.to_pandas(), args.out, "include_contigs.tsv")
        
        # Create a new contig_bin file
        uniquely_assigned_contigs = include_contigs_df\
            .filter(pl.col("confidence") == "high_confidence")\
            .drop("bin")\
            .rename({"assigned_bin": "bin"})\
            .select(["contig", "bin"])\
            .unique()

        new_contig_bins = data_processing.create_contig_bin_file(
            contig_bins=contig_bins.to_pandas(), 
            contamination= contamination.to_pandas(),
            include=uniquely_assigned_contigs.to_pandas()
        )
        data_processing.generate_output(new_contig_bins, args.out, "new_contig_bin.tsv")
        
    if args.write_bins:
        log.info("Write bins flag is set. Writing bins to file...")
        
        bin_dir = os.path.join(args.out, args.command + "_bins")
        
        data_processing.write_bins_from_contigs(new_contig_bins, assembly, bin_dir)
    
    log.info(f"Analysis Completed. Results are saved to: {args.out}")
    print("Analysis Completed. Results are saved to:", args.out)


from nanomotif.mtase_linker.dependencies import snakemake_create_environments, get_models, defensefinder_update, check_installation_MTase_linker
from nanomotif.mtase_linker.command import run_MTase_linker

def mtase_linker(args):
    if args.mtase_linker_command == "install":
        snakemake_create_environments(args)
        get_models(args)
        defensefinder_update(args)
        check_installation_MTase_linker(args)
    elif args.mtase_linker_command == "run":
        run_MTase_linker(args)
    else:
        print(f"Unknown MTase-linker command: {args.mtase_linker_command}")
       



def main():
    # Parse arguments
    parser = nm.argparser.create_parser()
    args = parser.parse_args()

    try: 
        os.environ["POLARS_MAX_THREADS"] = str(args.threads)
    except:
        os.environ["POLARS_MAX_THREADS"] = "1"
    import polars as pl
    if args.command in ["detect_contamination", "include_contigs", "MTase-linker"]:
        args.verbose = False
        args.seed = 1
    
    if args.command == "find_motifs":
        shared_setup(args, args.out)
        find_motifs(args, pl)
    elif args.command == "score_motifs":
        shared_setup(args, args.out)
        score_motifs(args, pl)
    elif args.command == "bin_consensus":
        shared_setup(args, args.out)
        bin_consensus(args, pl)
    elif args.command == "motif_discovery":
        shared_setup(args, args.out)
        motif_discovery(args, pl)

    elif args.command in ["detect_contamination", "include_contigs"]:
        shared_setup(args, args.out)
        binnary(args, pl)

    elif args.command == "MTase-linker":
        mtase_linker(args)

    elif args.command == "check_installation":
        args.out = "nanomotif_install_check"
        shared_setup(args, args.out)
        check_install(args, pl)

    else:
        parser.print_help()
        exit()
if __name__ == "__main__":
    main()



from nanomotif.constants import *
from nanomotif.candidate import Motif
from nanomotif.seq import DNAsequence, DNAarray
from nanomotif.evaluate import MotifSearcher
import nanomotif as nm
import pytest
from hypothesis import given, strategies as st
import itertools
import pandas as pd
import polars as pl
import polars.testing
import numpy as np


class TestMethylatedMotifOccourances:
    def test_known_motif_sequence(self):
        motif = nm.candidate.Motif('ACG', 0)
        sequence = 'TACGGACGCCACG'
        methylated_positions = np.array([1, 5])

        expected_methylated = np.array([1, 5])
        expected_nonmethylated = np.array([10])
        
        result = nm.evaluate.methylated_motif_occourances(motif, sequence, methylated_positions)
        np.testing.assert_array_equal(result[0], expected_methylated)
        np.testing.assert_array_equal(result[1], expected_nonmethylated)

    def test_known_motif_sequence_no_methylated(self):
        motif = nm.candidate.Motif('ACG', 0)
        sequence = 'TACGGACGCCACG'
        methylated_positions = np.array([])

        expected_methylated = np.array([])
        expected_nonmethylated = np.array([1, 5, 10])
        
        result = nm.evaluate.methylated_motif_occourances(motif, sequence, methylated_positions)
        np.testing.assert_array_equal(result[0], expected_methylated)
        np.testing.assert_array_equal(result[1], expected_nonmethylated)


class TestMotifSearcher:
    def test_init_with_valid_inputs(self):
        root_motif = Motif("ACGT", 0)
        contig_sequence = DNAsequence("ACGTACGTACGT")
        contig_sequences_sample = contig_sequence.sample_n_subsequences(
            2 * 2 + 1, 10
        )
        contig_pssm = contig_sequences_sample.pssm()
        contig_pileup = pl.DataFrame({"position": [1,2,3], "value": [0.1, 0.2, 0.3]})
        methylation_sequences = DNAarray(["ACGT", "CGTA", "GTAC"])
        padding = 2

        # Create MotifSearcher instance
        searcher = MotifSearcher(
            root_motif=root_motif,
            contig_sequence=contig_sequence,
            contig_pssm=contig_pssm,
            contig_pileup=contig_pileup,
            methylation_sequences=methylation_sequences,
            padding=padding
        )

        # Check that attributes are set correctly
        assert searcher.root_motif == root_motif
        assert searcher.contig_sequence == contig_sequence
        assert searcher.contig_pileup.equals(contig_pileup)
        assert np.array_equal(searcher.methylation_sequences, methylation_sequences)
        assert searcher.padding == padding

    def test_priority_function_zero_division_beta(self):
        root_motif = Motif("ACGT", 0)
        contig_sequence = DNAsequence("ACGTACGTACGT")
        contig_sequences_sample = contig_sequence.sample_n_subsequences(
            2 * 2 + 1, 10
        )
        contig_pssm = contig_sequences_sample.pssm()
        contig_pileup = pl.DataFrame({"position": [1,2,3], "value": [0.1, 0.2, 0.3]})
        methylation_sequences = DNAarray(["ACGT", "CGTA", "GTAC"])
        padding = 2

        # Create MotifSearcher instance
        searcher = MotifSearcher(
            root_motif=root_motif,
            contig_sequence=contig_sequence,
            contig_pssm=contig_pssm,
            contig_pileup=contig_pileup,
            methylation_sequences=methylation_sequences,
            padding=padding
        )

        class MockModel:
            def __init__(self, alpha, beta):
                self._alpha = alpha
                self._beta = beta

        next_model = MockModel(alpha=2, beta=3)
        root_model = MockModel(alpha=4, beta=0)

        priority = searcher._priority_function(next_model, root_model)

        expected_d_alpha = 1 - (next_model._alpha / root_model._alpha)  # 1 - (2/4) = 0.5
        expected_d_beta = 1  # ZeroDivisionError caught
        expected_priority = expected_d_alpha * expected_d_beta  # 0.5 * 1 = 0.5

        assert priority == pytest.approx(expected_priority)

    def test_scoring_function(self):
        root_motif = Motif("ACGT", 0)
        contig_sequence = DNAsequence("ACGTACGTACGT")
        contig_sequences_sample = contig_sequence.sample_n_subsequences(
            2 * 2 + 1, 10
        )
        contig_pssm = contig_sequences_sample.pssm()
        contig_pileup = pl.DataFrame({"position": [1,2,3], "value": [0.1, 0.2, 0.3]})
        methylation_sequences = DNAarray(["ACGT", "CGTA", "GTAC"])
        padding = 2

        # Create MotifSearcher instance
        searcher = MotifSearcher(
            root_motif=root_motif,
            contig_sequence=contig_sequence,
            contig_pssm=contig_pssm,
            contig_pileup=contig_pileup,
            methylation_sequences=methylation_sequences,
            padding=padding
        )

        class MockModel:
            def __init__(self, mean_value, std_dev):
                self._mean = mean_value
                self._std_dev = std_dev

            def mean(self):
                return self._mean

            def standard_deviation(self):
                return self._std_dev

        next_model = MockModel(mean_value=0.6, std_dev=0.1)
        current_model = MockModel(mean_value=0.4, std_dev=0.2)

        score = searcher._scoring_function(next_model, current_model)

        mean_diff = next_model.mean() - current_model.mean()  # 0.6 - 0.4 = 0.2
        expected_score = (next_model.mean() * -np.log10(next_model.standard_deviation())) * mean_diff
        expected_score = 0.6 * 1 * 0.2

        assert score == pytest.approx(expected_score)

class TestProcessSubpileup:
    def find_motifs(self):
        contig = "contig_3"
        min_kl_divergence = 0.1
        padding = 20
        minimum_methylation_fraction_confident = 0.75
        modtype = "a"

        pileup_path = nm.datasets.geobacillus_plasmids_pileup_path()
        pileup = nm.dataload.load_pileup(pileup_path, min_fraction=0.5, min_coverage=5)
        subpileup = pileup.pileup.filter(pl.col("mod_type") == modtype)
        assembly = nm.datasets.geobacillus_plasmids_assembly()

        result = nm.evaluate.process_subpileup(contig, modtype, subpileup, assembly, min_kl_divergence, padding, minimum_methylation_fraction_confident, False)
        return result

    def test_no_errors(self):
        self.find_motifs()
    
    def test_correct_output(self):
        result = self.find_motifs()
        expected_output = pl.DataFrame({
            "sequence":[
                "...................GATC..................", 
                "................ACCCA....................", 
                "................CCAAAT...................", 
                "...............G[AG].GAAG[TC].................."],
            "score":[1.97339, 1.399304, 1.388421, 0.572978],
            "contig":["contig_3", "contig_3", "contig_3", "contig_3"],
            "mod_type":["a", "a", "a", "a"],
            "alpha":[706, 65, 80, 84],
            "beta":[3, 1, 1, 1]
        })
        print(result)
        polars.testing.assert_frame_equal(result, expected_output)


class TestProcessSampleParallel():
    def run_function(self):
        pileup_path = nm.datasets.geobacillus_plasmids_pileup_path()
        pileup = nm.dataload.load_pileup(pileup_path, min_fraction=0.5, min_coverage=5)
        pileup = pileup.pileup
        assembly = nm.datasets.geobacillus_plasmids_assembly()

        return nm.evaluate.process_sample_parallel(assembly, pileup, threads=2, seed=1)
    
    def test_no_errors(self):
        self.run_function()

    def test_reprocuible_result(self):
        res1 = self.run_function().drop("model").sort(["contig", "mod_type", "motif"])
        res2 = self.run_function().drop("model").sort(["contig", "mod_type", "motif"])
        polars.testing.assert_frame_equal(res1, res2)




#%% Utility functions

#%% Reverse complement
def reverse_complement(seq):
    """
    Generate the reverse complement of a DNA sequence.

    Given a string representing a DNA sequence, this function returns its 
    reverse complement. The reverse complement is formed by first reversing 
    the original sequence and then taking the complement of each nucleotide. 
    This function recognizes standard DNA bases (A, T, G, C), ambiguous 
    nucleotides (R, Y, S, W, K, M, B, D, H, V, N), and special characters 
    (., [, and ]).

    Parameters:
    seq (str): A string representing the DNA sequence.

    Returns:
    str: The reverse complement of the given DNA sequence.

    Raises:
    KeyError: If the sequence contains characters not defined in the 
    complement mapping.

    Examples:
    >>> reverse_complement("ATGC")
    "GCAT"

    >>> reverse_complement("RYSWKM")
    "MKSWYR"

    Note:
    The function assumes that the input sequence is in uppercase. If the input
    contains lowercase characters, they will not be recognized and a KeyError
    will be raised.
    """

    complement = {
        "A": "T", "T": "A", "G": "C", "C": "G", 
        "N": "N", "R": "Y", "Y": "R", "S": "S", 
        "W": "W", "K": "M", "M": "K", "B": "V", 
        "D": "H", "H": "D", "V": "B", ".": ".",
        "[": "]", "]": "[", "N": "N"
    }

    return ''.join(complement[base] for base in reversed(seq))

#%% Function to predict motif type based on RM-type
def motif_type_predictions(df, subtype_column):
    """
    Adds a new column 'motif type' to the DataFrame based on predefined match conditions.

    Parameters:
    - df: MTase table.
    - subtype_column (str): A column which state the RM type of the MTase.

    Returns:
    - pandas.DataFrame: The modified DataFrame with a new 'motif type' column added.

    The function matches each row in the 'subtype_column' against predefined strings
    for four motif types: 'bipartite', 'palindromic', 'non-palindromic', and 'unknown'.
    It then assigns a corresponding label to each row in the new 'motif type' column.
    """

   
    match_string1 = ['Type_I']  
    match_string2 = ['Type_II'] 
    match_string3 = ['Type_III']  
    match_string4 = ['Type_IIG'] 
    output_string1 = 'bipartite'
    output_string2 = 'palindrome'
    output_string3 = 'non-palindrome'
    output_string4 = 'bipartite, non-palindrome'

    def classify_motif(x):
        if x in match_string1:
            return output_string1
        elif x in match_string2:
            return output_string2
        elif x in match_string3:
            return output_string3
        elif x in match_string4:
            return output_string4
        else:
            return 'undefined' 

    df['motif_type'] = df[subtype_column].apply(classify_motif)

    return df

#%% Function to convert HMM-hit string to RM subtype type
def RM_type_converter(df, gene_name_column):
    """
    Adds a new column 'sub_type' to the DataFrame based on predefined match conditions.

    Parameters:
    - df: MTase table.
    - subtype_column (str): A column which state the name of the HMM, which identified the MTase.

    Returns:
    - pandas.DataFrame: The modified DataFrame with a new 'sub_type' column added.

    The function matches each entry in the 'gene name' column against predefined strings
    for the HMM names: 'RM__Type_I_MTases', 'RM_Type_II__Type_II_MTases', 'RM_Type_IIG__Type_IIG', and 'RM_Type_III__Type_III_MTases'.
    It then assigns a corresponding RM type to each entry in the new 'sub_type' column.
    """

 
    match_string1 = ['RM__Type_I_MTases'] 
    match_string2 = ['RM_Type_II__Type_II_MTases']
    match_string3 = ['RM_Type_IIG__Type_IIG']
    match_string4 = ['RM_Type_III__Type_III_MTases']
    output_string1 = 'Type_I'
    output_string2 = 'Type_II'
    output_string3 = 'Type_IIG'
    output_string4 = 'Type_III'

    def classify_motif(x):
        if x in match_string1:
            return output_string1
        elif x in match_string2:
            return output_string2
        elif x in match_string3:
            return output_string3
        elif x in match_string4:
            return output_string4
        else:
            return 'undefined'  

    df['sub_type'] = df[gene_name_column].apply(classify_motif)

    return df

#%% Functions to predict mod type based on HMM hits

def mod_predictions_hmm(df, pfam_hmm_acc_column):
    """
    Creates a new column in the pfam hit table predicting mod type based on pfam family hits.

    Parameters:
    - df (pandas.DataFrame): pfam_hit_df.
    - pfam_hmm_acc_column: the pfam hmm accession number column.

    Returns:
    - Return the df with the new mod type column added.
    """

    a_interpro_acc = ["PF01555", "PF02384", "PF12161", "PF05869", "PF02086", "PF07669", "PF13651"]
    m_interpro_acc = ["PF00145"]

    def classify_mod(x):
        if x.split('.')[0] in a_interpro_acc:
            return "ac"
        elif x.split('.')[0] in m_interpro_acc:
            return "m"
        else:
            return None

    df['mod_type'] = df[pfam_hmm_acc_column].apply(classify_mod)

    return df
# %%
#Function to recode nanomotif bin-motifs modtype values.
def recode_mod_type(value):
    if value == 'm':
        return 'm'
    elif value == 'a' or value == '21839':
        return 'ac'
    else:
        return value